from ..draw import Mesh
from ..live2d import Live2D
from ..live2d_gl_wrapper import Live2DGLWrapper
from .draw_param import DrawParam


class FrameBufferObject:

    def __init__(self, fbo, rbo, tex):
        self.framebuffer = fbo
        self.renderbuffer = rbo
        self.texture = tex


class DrawParamOpenGL(DrawParam):

    def __init__(self):
        super().__init__()
        self.framebufferObject: FrameBufferObject | None = None
        self.shaderProgramOff = None
        self.textures = []
        self.transform = None
        self.gl = Live2DGLWrapper()
        self.firstDraw = True
        self.anisotropyExt = None
        self.maxAnisotropy = 0
        self.uvbo = None
        self.vbo = None
        self.ebo = None
        self.vertShader = None
        self.fragShader = None
        self.vertShaderOff = None
        self.fragShaderOff = None

    def getGL(self):
        return self.gl

    def setGL(self, aH):
        self.gl = aH

    def resize(self, ww, wh):
        self.gl.resize(ww, wh)

    def setTransform(self, aH):
        self.transform = aH

    def setupDraw(self):
        a_h = self.gl
        if self.firstDraw:
            self.initShader()
            self.firstDraw = False
            # self.anisotropyExt = aH.MAX_TEXTURE_MAX_ANISOTROPY
            # if self.anisotropyExt:
            #     self.maxAnisotropy = a_h.getParameter(self.anisotropyExt)

        a_h.disable(a_h.SCISSOR_TEST)
        a_h.disable(a_h.STENCIL_TEST)
        a_h.disable(a_h.DEPTH_TEST)
        a_h.frontFace(a_h.CW)
        a_h.enable(a_h.BLEND)
        a_h.colorMask(1, 1, 1, 1)
        a_h.bindBuffer(a_h.ARRAY_BUFFER, 0)
        a_h.bindBuffer(a_h.ELEMENT_ARRAY_BUFFER, 0)

    def drawTexture(self, texNo, screenColor, indexArray, vertexArray, uvArray, opacity, comp, multiplyColor):
        if opacity < 0.01 and self.clipBufPre_clipContextMask is None:
            return

        gl = self.gl

        a_w = self.baseRed * opacity
        a2 = self.baseGreen * opacity
        a5 = self.baseBlue * opacity
        a7 = self.baseAlpha * opacity
        if self.clipBufPre_clipContextMask is not None:
            # 根据蒙版进行裁剪
            gl.frontFace(gl.CCW)
            gl.useProgram(self.shaderProgram)
            self.vbo = bindOrCreateVBO(gl, self.vbo, vertexArray)
            self.ebo = bindOrCreateEBO(gl, self.ebo, indexArray)
            gl.vertexAttribPointer(self.a_position_Loc, 2, gl.FLOAT, False, 0, None)
            gl.enableVertexAttribArray(self.a_position_Loc)
            self.uvbo = bindOrCreateVBO(gl, self.uvbo, uvArray)
            gl.activeTexture(gl.TEXTURE1)
            gl.bindTexture(gl.TEXTURE_2D, self.textures[texNo])
            gl.uniform1i(self.s_texture0_Loc, 1)
            gl.vertexAttribPointer(self.a_texCoord_Loc, 2, gl.FLOAT, False, 0, None)
            gl.enableVertexAttribArray(self.a_texCoord_Loc)
            gl.uniformMatrix4fv(self.u_matrix_Loc, False, self.getClipBufPre_clipContextMask().matrixForMask)
            aY = self.getClipBufPre_clipContextMask().layoutChannelNo
            a4 = self.getChannelFlagAsColor(aY)
            gl.uniform4f(self.u_channelFlag, a4.r, a4.g, a4.b, a4.a)
            aI = self.getClipBufPre_clipContextMask().layoutBounds
            gl.uniform4f(self.u_baseColor_Loc, aI.x * 2.0 - 1.0, aI.y * 2.0 - 1.0, aI.getRight() * 2.0 - 1.0,
                         aI.getBottom() * 2.0 - 1.0)
            gl.uniform1i(self.u_maskFlag_Loc, True)
            gl.uniform4f(self.u_screenColor_Loc, *screenColor)
            gl.uniform4f(self.u_multiplyColor_Loc, *multiplyColor)
        elif self.clipBufPre_clipContextDraw is not None:
            # 绘制蒙版
            gl.useProgram(self.shaderProgramOff)
            self.vbo = bindOrCreateVBO(gl, self.vbo, vertexArray)
            self.ebo = bindOrCreateEBO(gl, self.ebo, indexArray)
            gl.enableVertexAttribArray(self.a_position_Loc_Off)
            gl.vertexAttribPointer(self.a_position_Loc_Off, 2, gl.FLOAT, False, 0, None)
            self.uvbo = bindOrCreateVBO(gl, self.uvbo, uvArray)
            gl.activeTexture(gl.TEXTURE1)
            gl.bindTexture(gl.TEXTURE_2D, self.textures[texNo])
            gl.uniform1i(self.s_texture0_Loc_Off, 1)
            gl.enableVertexAttribArray(self.a_texCoord_Loc_Off)
            gl.vertexAttribPointer(self.a_texCoord_Loc_Off, 2, gl.FLOAT, False, 0, None)
            gl.uniformMatrix4fv(self.u_clipMatrix_Loc_Off, False,
                                self.getClipBufPre_clipContextDraw().matrixForDraw)
            gl.uniformMatrix4fv(self.u_matrix_Loc_Off, False, self.matrix4x4)
            gl.activeTexture(gl.TEXTURE2)
            gl.bindTexture(gl.TEXTURE_2D, self.framebufferObject.texture)
            gl.uniform1i(self.s_texture1_Loc_Off, 2)
            aY = self.getClipBufPre_clipContextDraw().layoutChannelNo
            a4 = self.getChannelFlagAsColor(aY)
            gl.uniform4f(self.u_channelFlag_Loc_Off, a4.r, a4.g, a4.b, a4.a)
            gl.uniform4f(self.u_baseColor_Loc_Off, a_w, a2, a5, a7)
            gl.uniform4f(self.u_screenColor_Loc_Off, *screenColor)
            gl.uniform4f(self.u_multiplyColor_Loc_Off, *multiplyColor)
        else:
            gl.useProgram(self.shaderProgram)
            self.vbo = bindOrCreateVBO(gl, self.vbo, vertexArray)
            self.ebo = bindOrCreateEBO(gl, self.ebo, indexArray)
            gl.enableVertexAttribArray(self.a_position_Loc)
            gl.vertexAttribPointer(self.a_position_Loc, 2, gl.FLOAT, False, 0, None)
            self.uvbo = bindOrCreateVBO(gl, self.uvbo, uvArray)
            gl.activeTexture(gl.TEXTURE1)
            gl.bindTexture(gl.TEXTURE_2D, self.textures[texNo])
            gl.uniform1i(self.s_texture0_Loc, 1)
            gl.enableVertexAttribArray(self.a_texCoord_Loc)
            gl.vertexAttribPointer(self.a_texCoord_Loc, 2, gl.FLOAT, False, 0, None)
            gl.uniformMatrix4fv(self.u_matrix_Loc, False, self.matrix4x4)
            gl.uniform4f(self.u_baseColor_Loc, a_w, a2, a5, a7)
            gl.uniform1i(self.u_maskFlag_Loc, False)
            gl.uniform4f(self.u_screenColor_Loc, *screenColor)
            gl.uniform4f(self.u_multiplyColor_Loc, *multiplyColor)
        if self.culling:
            self.gl.enable(gl.CULL_FACE)
        else:
            self.gl.disable(gl.CULL_FACE)

        self.gl.enable(gl.BLEND)
        src_color = None
        src_factor = None
        dst_color = None
        dst_factor = None

        if comp == Mesh.COLOR_COMPOSITION_NORMAL:
            src_color = gl.ONE
            src_factor = gl.ONE_MINUS_SRC_ALPHA
            dst_color = gl.ONE
            dst_factor = gl.ONE_MINUS_SRC_ALPHA
        elif comp == Mesh.COLOR_COMPOSITION_SCREEN:
            src_color = gl.ONE
            src_factor = gl.ONE
            dst_color = gl.ZERO
            dst_factor = gl.ONE
        elif comp == Mesh.COLOR_COMPOSITION_MULTIPLY:
            src_color = gl.DST_COLOR
            src_factor = gl.ONE_MINUS_SRC_ALPHA
            dst_color = gl.ZERO
            dst_factor = gl.ONE

        gl.blendEquationSeparate(gl.FUNC_ADD, gl.FUNC_ADD)
        gl.blendFuncSeparate(src_color, src_factor, dst_color, dst_factor)
        # if self.anisotropyExt:
        #     g0.texParameteri(g0.TEXTURE_2D, self.anisotropyExt.TEXTURE_MAX_ANISOTROPY_EXT, self.maxAnisotropy)

        count = len(indexArray)
        gl.drawElements(gl.TRIANGLES, count, gl.UNSIGNED_SHORT, None)
        gl.bindTexture(gl.TEXTURE_2D, 0)

    def setTexture(self, aH, aI):
        size = len(self.textures)
        if aH >= size:
            for i in range(size, aH + 1):
                self.textures.append(None)
        self.textures[aH] = aI

    def initShader(self):
        aH = self.gl
        self.loadShaders2()
        self.a_position_Loc = aH.getAttribLocation(self.shaderProgram, "a_position")
        self.a_texCoord_Loc = aH.getAttribLocation(self.shaderProgram, "a_texCoord")
        self.u_matrix_Loc = aH.getUniformLocation(self.shaderProgram, "u_mvpMatrix")
        self.s_texture0_Loc = aH.getUniformLocation(self.shaderProgram, "s_texture0")
        self.u_channelFlag = aH.getUniformLocation(self.shaderProgram, "u_channelFlag")
        self.u_baseColor_Loc = aH.getUniformLocation(self.shaderProgram, "u_baseColor")
        self.u_maskFlag_Loc = aH.getUniformLocation(self.shaderProgram, "u_maskFlag")
        self.a_position_Loc_Off = aH.getAttribLocation(self.shaderProgramOff, "a_position")
        self.a_texCoord_Loc_Off = aH.getAttribLocation(self.shaderProgramOff, "a_texCoord")
        self.u_matrix_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "u_mvpMatrix")
        self.u_clipMatrix_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "u_clipMatrix")
        self.s_texture0_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "s_texture0")
        self.s_texture1_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "s_texture1")
        self.u_channelFlag_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "u_channelFlag")
        self.u_baseColor_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "u_baseColor")
        self.u_screenColor_Loc = aH.getUniformLocation(self.shaderProgram, "u_screenColor")
        self.u_screenColor_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "u_screenColor")
        self.u_multiplyColor_Loc = aH.getUniformLocation(self.shaderProgram, "u_multiplyColor")
        self.u_multiplyColor_Loc_Off = aH.getUniformLocation(self.shaderProgramOff, "u_multiplyColor")

    def disposeShader(self):
        aH = self.gl
        if self.shaderProgram:
            aH.deleteProgram(self.shaderProgram)
            self.shaderProgram = None

        if self.shaderProgramOff:
            aH.deleteProgram(self.shaderProgramOff)
            self.shaderProgramOff = None

    def compileShader(self, aJ, aN):
        aM = self.gl
        aL = aN
        aK = aM.createShader(aJ)
        if aK is None:
            print("_L0 to create shader")
            return None

        aM.shaderSource(aK, aL)
        aM.compileShader(aK)
        aH = aM.getShaderParameter(aK, aM.COMPILE_STATUS)
        if not aH:
            aI = aM.getShaderInfoLog(aK)
            print(f"_L0 to compile shader : {aI}")
            aM.deleteShader(aK)
            return None

        return aK

    def loadShaders2(self):
        aN = self.gl
        self.shaderProgram = aN.createProgram()
        if not self.shaderProgram:
            return False

        self.shaderProgramOff = aN.createProgram()
        if not self.shaderProgramOff:
            return False

        aK = ("#version 120\n"
              "attribute vec2 a_position;"
              "attribute vec2 a_texCoord;"
              "varying vec2 v_texCoord;"
              "varying vec4 v_clipPos;"
              "uniform mat4 u_mvpMatrix;"
              "void main(){"
              "    gl_Position = u_mvpMatrix * vec4(a_position, 0.0, 1.0);"
              "    v_clipPos = gl_Position;"
              "    v_texCoord = a_texCoord;"
              "    v_texCoord.y = 1.0 - v_texCoord.y;}")
        aM = ("#version 120\n"
              "varying vec2       v_texCoord;"
              "varying vec4       v_clipPos;"
              "uniform sampler2D  s_texture0;"
              "uniform vec4       u_channelFlag;"
              "uniform vec4       u_baseColor;"
              "uniform bool       u_maskFlag;"
              "uniform vec4       u_screenColor;"
              "uniform vec4       u_multiplyColor;"
              "void main(){"
              "    vec4 smpColor;"
              "    if(u_maskFlag){"
              "        float isInside = "
              "            step(u_baseColor.x, v_clipPos.x/v_clipPos.w)"
              "          * step(u_baseColor.y, v_clipPos.y/v_clipPos.w)"
              "          * step(v_clipPos.x/v_clipPos.w, u_baseColor.z)"
              "          * step(v_clipPos.y/v_clipPos.w, u_baseColor.w);"
              "        smpColor = u_channelFlag * texture2D(s_texture0, v_texCoord).a * isInside;"
              "    }else{"
              "        smpColor = texture2D(s_texture0 , v_texCoord);"
              "        smpColor.rgb = smpColor.rgb * u_multiplyColor.rgb;"
              "        smpColor.rgb = smpColor.rgb + u_screenColor.rgb - (smpColor.rgb * u_screenColor.rgb);"
              "        smpColor = smpColor * u_baseColor;"
              "        smpColor = vec4(smpColor.rgb * smpColor.a, smpColor.a);"
              "    }"
              "    gl_FragColor = smpColor;}")
        aL = ("#version 120\n"
              "attribute vec2     a_position;"
              "attribute vec2     a_texCoord;"
              "varying vec2       v_texCoord;"
              "varying vec4       v_clipPos;"
              "uniform mat4       u_mvpMatrix;"
              "uniform mat4       u_clipMatrix;"
              "void main(){"
              "    vec4 pos = vec4(a_position, 0, 1.0);"
              "    gl_Position = u_mvpMatrix * pos;"
              "    v_clipPos = u_clipMatrix * pos;"
              "    v_texCoord = a_texCoord;"
              "    v_texCoord.y = 1.0 - v_texCoord.y;}")
        aJ = ("#version 120\n"
              "varying   vec2   v_texCoord;"
              "varying   vec4   v_clipPos;"
              "uniform sampler2D  s_texture0;"
              "uniform sampler2D  s_texture1;"
              "uniform vec4       u_channelFlag;"
              "uniform vec4       u_baseColor;"
              "uniform vec4       u_screenColor;"
              "uniform vec4       u_multiplyColor;"
              "void main(){"
              "    vec4 col_formask = texture2D(s_texture0, v_texCoord);"
              "    col_formask.rgb = col_formask.rgb * u_multiplyColor.rgb;"
              "    col_formask.rgb = col_formask.rgb + u_screenColor.rgb - (col_formask.rgb * u_screenColor.rgb);"
              "    col_formask = col_formask * u_baseColor;"
              "    col_formask.rgb = col_formask.rgb * col_formask.a;"
              "    vec4 clipMask = texture2D(s_texture1, v_clipPos.xy / v_clipPos.w) * u_channelFlag;"
              "    float maskVal = clipMask.r + clipMask.g + clipMask.b + clipMask.a;"
              "    col_formask = col_formask * maskVal;"
              "    gl_FragColor = col_formask;}")
        self.vertShader = self.compileShader(aN.VERTEX_SHADER, aK)
        if not self.vertShader:
            print("Vertex shader compile li_!")
            return False

        self.vertShaderOff = self.compileShader(aN.VERTEX_SHADER, aL)
        if not self.vertShaderOff:
            print("OffVertex shader compile li_!")
            return False

        self.fragShader = self.compileShader(aN.FRAGMENT_SHADER, aM)
        if not self.fragShader:
            print("Fragment shader compile li_!")
            return False

        self.fragShaderOff = self.compileShader(aN.FRAGMENT_SHADER, aJ)
        if not self.fragShaderOff:
            print("OffFragment shader compile li_!")
            return False

        aN.attachShader(self.shaderProgram, self.vertShader)
        aN.attachShader(self.shaderProgram, self.fragShader)
        aN.attachShader(self.shaderProgramOff, self.vertShaderOff)
        aN.attachShader(self.shaderProgramOff, self.fragShaderOff)
        aN.linkProgram(self.shaderProgram)
        aN.linkProgram(self.shaderProgramOff)
        aH = aN.getProgramParameter(self.shaderProgram, aN.LINK_STATUS)
        aX = aN.getProgramParameter(self.shaderProgramOff, aN.LINK_STATUS)
        if not aH or not aX:
            if aH:
                aI = aN.getProgramInfoLog(self.shaderProgram)
            else:
                aI = aN.getProgramInfoLog(self.shaderProgramOff)
            print(f"failed to link program: {aI}")
            if self.vertShader:
                aN.deleteShader(self.vertShader)
                self.vertShader = 0

            if self.fragShader:
                aN.deleteShader(self.fragShader)
                self.fragShader = 0

            if self.shaderProgram:
                aN.deleteProgram(self.shaderProgram)
                self.shaderProgram = 0

            if self.vertShaderOff:
                aN.deleteShader(self.vertShaderOff)
                self.vertShaderOff = 0

            if self.fragShaderOff:
                aN.deleteShader(self.fragShaderOff)
                self.fragShaderOff = 0

            if self.shaderProgramOff:
                aN.deleteProgram(self.shaderProgramOff)
                self.shaderProgramOff = 0

            return False

        return True

    def createFramebuffer(self):
        aL = self.gl
        aK = Live2D.clippingMaskBufferSize
        aJ = aL.createFramebuffer()
        aL.bindFramebuffer(aL.FRAMEBUFFER, aJ)
        aH = aL.createRenderbuffer()
        aL.bindRenderbuffer(aL.RENDERBUFFER, aH)
        aL.renderbufferStorage(aL.RENDERBUFFER, aL.RGBA4, aK, aK)
        aL.framebufferRenderbuffer(aL.FRAMEBUFFER, aL.COLOR_ATTACHMENT0, aL.RENDERBUFFER, aH)
        aI = aL.createTexture()
        aL.bindTexture(aL.TEXTURE_2D, aI)
        aL.texImage2D(aL.TEXTURE_2D, 0, aL.RGBA, aK, aK, 0, aL.RGBA, aL.UNSIGNED_BYTE, None)
        aL.texParameteri(aL.TEXTURE_2D, aL.TEXTURE_MIN_FILTER, aL.LINEAR)
        aL.texParameteri(aL.TEXTURE_2D, aL.TEXTURE_MAG_FILTER, aL.LINEAR)
        aL.texParameteri(aL.TEXTURE_2D, aL.TEXTURE_WRAP_S, aL.CLAMP_TO_EDGE)
        aL.texParameteri(aL.TEXTURE_2D, aL.TEXTURE_WRAP_T, aL.CLAMP_TO_EDGE)
        aL.framebufferTexture2D(aL.FRAMEBUFFER, aL.COLOR_ATTACHMENT0, aL.TEXTURE_2D, aI, 0)
        aL.bindTexture(aL.TEXTURE_2D, 0)
        aL.bindRenderbuffer(aL.RENDERBUFFER, 0)
        aL.bindFramebuffer(aL.FRAMEBUFFER, 0)
        self.framebufferObject = FrameBufferObject(
            fbo=aJ,
            rbo=aH,
            tex=aI
        )


def bindOrCreateVBO(gl, vbo, data):
    if vbo is None:
        vbo = gl.createBuffer()

    gl.bindBuffer(gl.ARRAY_BUFFER, vbo)
    gl.bufferData(gl.ARRAY_BUFFER, data, gl.DYNAMIC_DRAW)
    return vbo


def bindOrCreateEBO(gl, ebo, data):
    if ebo is None:
        ebo = gl.createBuffer()

    gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, ebo)
    gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, data, gl.DYNAMIC_DRAW)
    return ebo



from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from .part import PartsData

class PartsDataContext:

    def __init__(self, parts):
        super().__init__()
        self.partsOpacity = None
        self.partsData: Optional['PartsData'] = parts
        self.screenColor = [0.0, 0.0, 0.0, 1.0]
        self.multiplyColor = [1.0, 1.0, 1.0, 1.0]

    def getPartsOpacity(self):
        return self.partsOpacity

    def setPartsOpacity(self, value):
        self.partsOpacity = value

    def setPartScreenColor(self, r, g, b, a):
        self.screenColor[0] = r
        self.screenColor[1] = g
        self.screenColor[2] = b
        self.screenColor[3] = a

    def setPartMultiplyColor(self, r, g, b, a):
        self.multiplyColor[0] = r
        self.multiplyColor[1] = g
        self.multiplyColor[2] = b
        self.multiplyColor[3] = a



import random

from ...core import UtSystem


class L2DEyeBlink:

    def __init__(self):
        self.nextBlinkTime = None
        self.stateStartTime = None
        self.blinkIntervalMsec = None
        self.eyeState = EyeState.STATE_FIRST
        self.blinkIntervalMsec = 4000
        self.closingMotionMsec = 100
        self.closedMotionMsec = 50
        self.openingMotionMsec = 150
        self.closeIfZero = True
        self.eyeID_L = "PARAM_EYE_L_OPEN"
        self.eyeID_R = "PARAM_EYE_R_OPEN"

    def calcNextBlink(self):
        time = UtSystem.getUserTimeMSec()
        r = random.random()
        return time + r * (2 * self.blinkIntervalMsec - 1)

    def setInterval(self, blinkIntervalMsec):
        self.blinkIntervalMsec = blinkIntervalMsec

    def setEyeMotion(self, closingMotionMsec, closedMotionMsec, openingMotionMsec):
        self.closingMotionMsec = closingMotionMsec
        self.closedMotionMsec = closedMotionMsec
        self.openingMotionMsec = openingMotionMsec

    def updateParam(self, model):
        time = UtSystem.getUserTimeMSec()
        eye_param_value = 0

        _switch_2297 = self.eyeState
        if self.eyeState == EyeState.STATE_CLOSING:
            t = (time - self.stateStartTime) / self.closingMotionMsec
            if t >= 1:
                t = 1
                self.eyeState = EyeState.STATE_CLOSED
                self.stateStartTime = time

            eye_param_value = 1 - t
        elif self.eyeState == EyeState.STATE_CLOSED:
            t = (time - self.stateStartTime) / self.closedMotionMsec
            if t >= 1:
                self.eyeState = EyeState.STATE_OPENING
                self.stateStartTime = time

            eye_param_value = 0
        elif self.eyeState == EyeState.STATE_OPENING:
            t = (time - self.stateStartTime) / self.openingMotionMsec
            if t >= 1:
                t = 1
                self.eyeState = EyeState.STATE_INTERVAL
                self.nextBlinkTime = self.calcNextBlink()

            eye_param_value = t
        elif self.eyeState == EyeState.STATE_INTERVAL:
            if self.nextBlinkTime < time:
                self.eyeState = EyeState.STATE_CLOSING
                self.stateStartTime = time
            eye_param_value = 1
        else:
            self.eyeState = EyeState.STATE_INTERVAL
            self.nextBlinkTime = self.calcNextBlink()
            eye_param_value = 1
        if not self.closeIfZero:
            eye_param_value = -eye_param_value
        model.setParamFloat(self.eyeID_L, eye_param_value)
        model.setParamFloat(self.eyeID_R, eye_param_value)


class EyeState:
    STATE_FIRST = "STATE_FIRST"
    STATE_INTERVAL = "STATE_INTERVAL"
    STATE_CLOSING = "STATE_CLOSING"
    STATE_CLOSED = "STATE_CLOSED"
    STATE_OPENING = "STATE_OPENING"



from abc import ABC

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from live2d.core.deformer import Deformer


class DeformerContext(ABC):

    def __init__(self, deformer: 'Deformer'):
        self.partsIndex = None
        self.outsideParam = False
        self.available = True
        self.deformer = deformer
        self.totalScale = 1.0
        self.interpolatedOpacity = 1.0
        self.totalOpacity = 1.0

    def isAvailable(self) -> bool:
        return self.available and not self.outsideParam

    def setAvailable(self, value: bool):
        self.available = value

    def getDeformer(self):
        return self.deformer

    def setPartsIndex(self, aH):
        self.partsIndex = aH

    def getPartsIndex(self):
        return self.partsIndex

    def isOutsideParam(self):
        return self.outsideParam

    def setOutsideParam(self, value):
        self.outsideParam = value

    def getTotalScale(self):
        return self.totalScale

    def setTotalScale_notForClient(self, aH):
        self.totalScale = aH

    def getInterpolatedOpacity(self):
        return self.interpolatedOpacity

    def setInterpolatedOpacity(self, value: float):
        self.interpolatedOpacity = value

    def getTotalOpacity(self):
        return self.totalOpacity

    def setTotalOpacity(self, aH):
        self.totalOpacity = aH



from ..Live2DFramework import Live2DFramework
from .l2d_parts_param import L2DPartsParam
from ...core import Array, UtSystem


class L2DPose:

    def __init__(self):
        self.lastTime = 0
        self.lastModel = None
        self.partsGroups = Array()

    def updateParam(self, model):
        if model != self.lastModel:
            self.initParam(model)

        self.lastModel = model
        cur_time = UtSystem.getUserTimeMSec()
        delta_time_sec = (0 if (self.lastTime == 0) else (cur_time - self.lastTime) / 1000.0)
        self.lastTime = cur_time
        if delta_time_sec < 0:
            delta_time_sec = 0
        for i in range(len(self.partsGroups)):
            self.normalizePartsOpacityGroup(model, self.partsGroups[i], delta_time_sec)
            self.copyOpacityOtherParts(model, self.partsGroups[i])

    def initParam(self, model):
        for i in range(len(self.partsGroups)):
            parts_group = self.partsGroups[i]
            for j in range(len(parts_group)):
                parts_group[j].initIndex(model)
                parts_index = parts_group[j].partsIndex
                param_index = parts_group[j].paramIndex
                if parts_index < 0:
                    continue
                v = (model.getParamFloat(param_index) != 0)
                model.setPartsOpacity(parts_index, (1.0 if v else 0.0))
                model.setParamFloat(param_index, (1.0 if v else 0.0))
                if parts_group[j].link is None:
                    continue
                for k in range(len(parts_group[j].link)):
                    parts_group[j].link[k].initIndex(model)

    @staticmethod
    def normalizePartsOpacityGroup(model, partsGroup, deltaTimeSec):
        visible_parts = -1
        visible_opacity = 1.0
        clear_time_sec = 0.5
        phi = 0.5
        max_back_opacity = 0.15
        for i in range(len(partsGroup)):
            parts_index = partsGroup[i].partsIndex
            param_index = partsGroup[i].paramIndex
            if parts_index < 0:
                continue
            if model.getParamFloat(param_index) != 0:
                if visible_parts >= 0:
                    break

                visible_parts = i
                visible_opacity = model.getPartsOpacity(parts_index)
                visible_opacity += deltaTimeSec / clear_time_sec
                if visible_opacity > 1:
                    visible_opacity = 1

        if visible_parts < 0:
            visible_parts = 0
            visible_opacity = 1

        for i in range(len(partsGroup)):
            parts_index = partsGroup[i].partsIndex
            if parts_index < 0:
                continue
            if visible_parts == i:
                model.setPartsOpacity(parts_index, visible_opacity)
            else:
                opacity = model.getPartsOpacity(parts_index)
                if visible_opacity < phi:
                    a1 = visible_opacity * (phi - 1) / phi + 1
                else:
                    a1 = (1 - visible_opacity) * phi / (1 - phi)

                back_op = (1 - a1) * (1 - visible_opacity)
                if back_op > max_back_opacity:
                    a1 = 1 - max_back_opacity / (1 - visible_opacity)

                opacity = min(opacity, a1)

                model.setPartsOpacity(parts_index, opacity)

    @staticmethod
    def copyOpacityOtherParts(model, partsGroup):
        for i_group in range(len(partsGroup)):
            parts_param = partsGroup[i_group]
            if parts_param.link is None:
                continue
            if parts_param.partsIndex < 0:
                continue
            opacity = model.getPartsOpacity(parts_param.partsIndex)
            for link_parts in parts_param.link:
                if link_parts.partsIndex < 0:
                    continue
                model.setPartsOpacity(link_parts.partsIndex, opacity)

    @staticmethod
    def load(buf):
        ret = L2DPose()
        pm = Live2DFramework.getPlatformManager()
        json = pm.jsonParseFromBytes(buf)
        pose_list_info = json.get("parts_visible")
        pose_num = len(pose_list_info)
        for i_pose in range(pose_num):
            pose_info = pose_list_info[i_pose]
            id_list_info = pose_info.get("group")
            id_num = len(id_list_info)
            parts_group = Array()
            for i_group in range(id_num):
                parts_info = id_list_info[i_group]
                parts = L2DPartsParam(parts_info["id"])
                parts_group.append(parts)
                if parts_info.get("link") is None:
                    continue
                link_list_info = parts_info.get("link")
                link_num = len(link_list_info)
                parts.link = Array()
                for i_link in range(link_num):
                    link_parts = L2DPartsParam(link_list_info[i_link])
                    parts.link.append(link_parts)

            ret.partsGroups.append(parts_group)

        return ret



from ...core import PhysicsHair, Array, UtSystem
from ..Live2DFramework import Live2DFramework


class L2DPhysics:

    def __init__(self):
        self.physicsList = Array()
        self.startTimeMSec = UtSystem.getUserTimeMSec()

    def updateParam(self, model):
        time_m_sec = UtSystem.getUserTimeMSec() - self.startTimeMSec
        for i in range(len(self.physicsList)):
            self.physicsList[i].update(model, time_m_sec)

    @staticmethod
    def load(buf):
        ret = L2DPhysics()
        pm = Live2DFramework.getPlatformManager()
        json = pm.jsonParseFromBytes(buf)
        params = json.get("physics_hair")
        param_num = len(params)
        for i in range(param_num):
            param = params[i]
            physics = PhysicsHair()
            setup = param.get("setup")
            length = float(len(setup))
            resist = float(setup.get("regist"))
            mass = float(setup.get("mass"))
            physics.setup(length, resist, mass)
            src_list = param.get("src")
            src_num = len(src_list)
            for j in range(src_num):
                src = src_list[j]
                tid = src.get("id")
                type_str = src.get("ptype")
                if type_str == "x":
                    t = PhysicsHair.SRC_TO_X
                elif type_str == "y":
                    t = PhysicsHair.SRC_TO_Y
                elif type_str == "angle":
                    t = PhysicsHair.SRC_TO_G_ANGLE
                else:
                    raise Exception("error")

                scale = float(src.get("scale"))
                weight = float(src.get("weight"))
                physics.addSrcParam(t, tid, scale, weight)

            target_list = param.get("targets")
            target_num = len(target_list)
            for j in range(target_num):
                target = target_list[j]
                tid = target.get("id")
                type_str = target.get("ptype")
                if type_str == "angle":
                    t = PhysicsHair.TARGET_FROM_ANGLE
                elif type_str == "angle_v":
                    t = PhysicsHair.TARGET_FROM_ANGLE_V
                else:
                    raise Exception("live2d", "Invalid parameter:PhysicsHair.Target")

                scale = float(target.get("scale"))
                weight = float(target.get("weight"))
                physics.addTargetParam(t, tid, scale, weight)

            ret.physicsList.append(physics)

        return ret



from .deformer import Deformer
from .deformer_context import DeformerContext


class WarpContext(DeformerContext):

    def __init__(self, dfm):
        super().__init__(dfm)

        self.tmpDeformerIndex = Deformer.DEFORMER_INDEX_NOT_INIT
        self.interpolatedPoints = None
        self.transformedPoints = None



import json
from typing import Any, Dict

import OpenGL.GL as gl
from PIL import Image

from .core import Live2DModelOpenGL


class PlatformManager:

    def loadBytes(self, path) -> bytes:
        with open(path, 'rb') as f:
            return f.read()

    def loadLive2DModel(self, path) -> Live2DModelOpenGL:
        with open(path, 'rb') as f:
            return Live2DModelOpenGL.loadModel(f.read())

    def loadTexture(self, live2DModel, no, path):
        image = Image.open(path)
        if image.mode != 'RGBA':
            image = image.convert("RGBA")
        image_data = image.tobytes()
        width, height = image.size
        gl.glEnable(gl.GL_TEXTURE_2D)
        texture = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
        gl.glTexImage2D(
            gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, width, height, 0, gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, image_data
        )
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR_MIPMAP_NEAREST)
        gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        gl.glGenerateMipmap(gl.GL_TEXTURE_2D)
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
        live2DModel.setTexture(no, texture)

    def jsonParseFromBytes(self, path) -> Dict[str, Any]:
        return json.loads(path)



from .l2d_expression_motion import L2DExpressionMotion
from .l2d_expression_param import L2DExpressionParam
from .l2d_motion_manager import L2DMotionManager
from .l2d_target_point import L2DTargetPoint
from .l2d_eye_blink import L2DEyeBlink


class Id:
    __instances = {}

    def __init__(self, aH: str):
        self.id: str = aH

    def __str__(self) -> str:
        return self.id

    def __eq__(self, other) -> bool:
        if isinstance(other, Id):
            return id(other) == id(self) or other.id == self.id
        elif isinstance(other, str):
            return other == self.id
        return False

    @staticmethod
    def DST_BASE_ID() -> 'Id':
        return Id.getID("DST_BASE")

    @staticmethod
    def getID(idStr: str) -> 'Id':
        if not isinstance(idStr, str):
            raise RuntimeError
        id_obj = Id.__instances.get(idStr, None)
        if id_obj is None:
            id_obj = Id(idStr)
            Id.__instances[idStr] = id_obj
        return id_obj

    @staticmethod
    def releaseStored() -> None:
        Id.__instances.clear()



from typing import List, Optional

from .clip_context import ClipContext
from .clip_matrix import ClipMatrix
from .clip_rectf import ClipRectF
from .texture_info import TextureInfo
from ..DEF import VERTEX_STEP, VERTEX_OFFSET
from ..live2d import Live2D
from ..type import Array, Float32Array


class ClippingManagerOpenGL:
    CHANNEL_COUNT = 4

    def __init__(self, aJ):
        self.clipContextList: List[Optional[ClipContext]] = Array()
        self.glcontext = aJ.gl
        self.dpGL = aJ
        self.curFrameNo = 0
        self.firstError_clipInNotUpdate = True
        self.colorBuffer = 0
        self.isInitGLFBFunc: bool = False
        self.tmpBoundsOnModel = ClipRectF()
        self.genMaskRenderTexture()
        self.tmpModelToViewMatrix = ClipMatrix()
        self.tmpMatrix2 = ClipMatrix()
        self.tmpMatrixForMask = ClipMatrix()
        self.tmpMatrixForDraw = ClipMatrix()
        self.channelColors: List[Optional[TextureInfo]] = Array()
        aI = TextureInfo()
        aI.r = 0
        aI.g = 0
        aI.b = 0
        aI.a = 1
        self.channelColors.append(aI)
        aI = TextureInfo()
        aI.r = 1
        aI.g = 0
        aI.b = 0
        aI.a = 0
        self.channelColors.append(aI)
        aI = TextureInfo()
        aI.r = 0
        aI.g = 1
        aI.b = 0
        aI.a = 0
        self.channelColors.append(aI)
        aI = TextureInfo()
        aI.r = 0
        aI.g = 0
        aI.b = 1
        aI.a = 0
        self.channelColors.append(aI)
        for aH in range(0, len(self.channelColors), 1):
            self.dpGL.setChannelFlagAsColor(aH, self.channelColors[aH])

    def releaseShader(self):
        aI = len(Live2D.frameBuffers)
        for aH in range(0, aI, 1):
            self.dpGL.deleteFramebuffer(Live2D.frameBuffers[aH].framebuffer)

        Live2D.frameBuffers = []
        Live2D.__glContext = []

    def init(self, aO, aN, aL):
        for aM in range(0, len(aN), 1):
            aH = aN[aM].getClipIDList()
            if aH is None:
                continue

            aJ = self.findSameClip(aH)
            if aJ is None:
                aJ = ClipContext(self, aO, aH)
                self.clipContextList.append(aJ)

            aI = aN[aM].getId()
            aK = aO.getDrawDataIndex(aI)
            aJ.addClippedDrawData(aI, aK)
            aP = aL[aM]
            aP.clipBufPre_clipContext = aJ

    def genMaskRenderTexture(self):
        self.dpGL.createFramebuffer()

    def setupClip(self, a1, aQ):
        aK = 0
        for aO in range(0, len(self.clipContextList), 1):
            aP = self.clipContextList[aO]
            self.calcClippedDrawTotalBounds(a1, aP)
            if aP.isUsing:
                aK += 1

        if aK > 0:
            oldFbo = aQ.gl.getParameter(aQ.gl.FRAMEBUFFER_BINDING)
            rect = Array(4)
            rect[0] = 0
            rect[1] = 0
            rect[2] = aQ.gl.width
            rect[3] = aQ.gl.height
            aQ.gl.viewport(0, 0, Live2D.clippingMaskBufferSize, Live2D.clippingMaskBufferSize)
            self.setupLayoutBounds(aK)
            aQ.gl.bindFramebuffer(aQ.gl.FRAMEBUFFER, aQ.framebufferObject.framebuffer)
            aQ.gl.clearColor(0, 0, 0, 0)
            aQ.gl.clear(aQ.gl.COLOR_BUFFER_BIT)
            for aO in range(0, len(self.clipContextList), 1):
                aP = self.clipContextList[aO]
                aT = aP.allClippedDrawRect
                aV = aP.layoutBounds
                aJ = 0.05
                self.tmpBoundsOnModel.setRect(aT)
                self.tmpBoundsOnModel.expand(aT.width * aJ, aT.height * aJ)
                aZ = aV.width / self.tmpBoundsOnModel.width
                aY = aV.height / self.tmpBoundsOnModel.height
                self.tmpMatrix2.identity()
                self.tmpMatrix2.translate(-1, -1, 0)
                self.tmpMatrix2.scale(2, 2, 1)
                self.tmpMatrix2.translate(aV.x, aV.y, 0)
                self.tmpMatrix2.scale(aZ, aY, 1)
                self.tmpMatrix2.translate(-self.tmpBoundsOnModel.x, -self.tmpBoundsOnModel.y, 0)
                self.tmpMatrixForMask.setMatrix(self.tmpMatrix2.m)
                self.tmpMatrix2.identity()
                self.tmpMatrix2.translate(aV.x, aV.y, 0)
                self.tmpMatrix2.scale(aZ, aY, 1)
                self.tmpMatrix2.translate(-self.tmpBoundsOnModel.x, -self.tmpBoundsOnModel.y, 0)
                self.tmpMatrixForDraw.setMatrix(self.tmpMatrix2.m)
                aH = self.tmpMatrixForMask.getArray()
                for aX in range(0, 16, 1):
                    aP.matrixForMask[aX] = aH[aX]

                a0 = self.tmpMatrixForDraw.getArray()
                for aX in range(0, 16, 1):
                    aP.matrixForDraw[aX] = a0[aX]

                aS = len(aP.clippingMaskDrawIndexList)
                for aU in range(0, aS, 1):
                    aR = aP.clippingMaskDrawIndexList[aU]
                    aI = a1.getDrawData(aR)
                    aL = a1.getDrawContext(aR)
                    aQ.setClipBufPre_clipContextForMask(aP)
                    aI.draw(aQ, a1, aL)

            aQ.gl.bindFramebuffer(aQ.gl.FRAMEBUFFER, oldFbo)
            aQ.setClipBufPre_clipContextForMask(None)
            aQ.gl.viewport(rect[0], rect[1], rect[2], rect[3])

    def getColorBuffer(self):
        return self.colorBuffer

    def findSameClip(self, aK):
        for aN in range(0, len(self.clipContextList), 1):
            aO = self.clipContextList[aN]
            aH = len(aO.clipIDList)
            if aH != len(aK):
                continue

            aI = 0
            for aM in range(0, aH, 1):
                aL = aO.clipIDList[aM]
                for aJ in range(0, aH, 1):
                    if aK[aJ] == aL:
                        aI += 1
                        break

            if aI == aH:
                return aO

        return None

    def calcClippedDrawTotalBounds(self, a6, aV):
        aU = a6.model.getModelImpl().getCanvasWidth()
        a5 = a6.model.getModelImpl().getCanvasHeight()
        aJ = aU if aU > a5 else a5
        aT = aJ
        aR = aJ
        aS = 0
        aP = 0
        aL = len(aV.clippedDrawContextList)
        for aM in range(0, aL, 1):
            aW = aV.clippedDrawContextList[aM]
            aN = aW.drawDataIndex
            aK = a6.getDrawContext(aN)
            if aK.isAvailable():
                aX = aK.getTransformedPoints()
                a4 = len(aX)
                aI = Float32Array(a4)
                aH = Float32Array(a4)
                aO = 0
                for a3 in range(VERTEX_OFFSET, a4, VERTEX_STEP):
                    aI[aO] = aX[a3]
                    aH[aO] = aX[a3 + 1]
                    aO += 1

                a2 = min(aI)
                a1 = min(aH)
                a0 = max(aI)
                aZ = max(aH)
                if a2 < aT:
                    aT = a2

                if a1 < aR:
                    aR = a1

                if a0 > aS:
                    aS = a0

                if aZ > aP:
                    aP = aZ

        if aT == aJ:
            aV.allClippedDrawRect.x = 0
            aV.allClippedDrawRect.y = 0
            aV.allClippedDrawRect.width = 0
            aV.allClippedDrawRect.height = 0
            aV.isUsing = False
        else:
            aQ = aS - aT
            aY = aP - aR
            aV.allClippedDrawRect.x = aT
            aV.allClippedDrawRect.y = aR
            aV.allClippedDrawRect.width = aQ
            aV.allClippedDrawRect.height = aY
            aV.isUsing = True

    def setupLayoutBounds(self, aQ):
        aI = aQ / ClippingManagerOpenGL.CHANNEL_COUNT
        aP = aQ % ClippingManagerOpenGL.CHANNEL_COUNT
        aI = int(aI)
        aP = int(aP)
        aH = 0
        for aJ in range(0, ClippingManagerOpenGL.CHANNEL_COUNT, 1):
            aM = aI + (1 if aJ < aP else 0)
            if aM == 1:
                aL = self.clipContextList[aH]
                aH += 1
                aL.layoutChannelNo = aJ
                aL.layoutBounds.x = 0
                aL.layoutBounds.y = 0
                aL.layoutBounds.width = 1
                aL.layoutBounds.height = 1
            elif aM == 2:
                for aO in range(0, aM, 1):
                    aN = aO % 2
                    aN = int(aN)
                    aL = self.clipContextList[aH]
                    aH += 1
                    aL.layoutChannelNo = aJ
                    aL.layoutBounds.x = aN * 0.5
                    aL.layoutBounds.y = 0
                    aL.layoutBounds.width = 0.5
                    aL.layoutBounds.height = 1
            elif aM <= 4:
                for aO in range(0, aM, 1):
                    aN = aO % 2
                    aK = aO / 2
                    aN = int(aN)
                    aK = int(aK)
                    aL = self.clipContextList[aH]
                    aH += 1
                    aL.layoutChannelNo = aJ
                    aL.layoutBounds.x = aN * 0.5
                    aL.layoutBounds.y = aK * 0.5
                    aL.layoutBounds.width = 0.5
                    aL.layoutBounds.height = 0.5
            elif aM <= 9:
                for aO in range(0, aM, 1):
                    aN = aO % 3
                    aK = aO / 3
                    aN = int(aN)
                    aK = int(aK)
                    aL = self.clipContextList[aH]
                    aH += 1
                    aL.layoutChannelNo = aJ
                    aL.layoutBounds.x = aN / 3
                    aL.layoutBounds.y = aK / 3
                    aL.layoutBounds.width = 1 / 3
                    aL.layoutBounds.height = 1 / 3



import time

RESET = "\033[0m"
RED = "\033[31m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
BLUE = "\033[34m"
MAGENTA = "\033[35m"
CYAN = "\033[36m"
WHITE = "\033[37m"

__enable = True


def setLogEnable(v: bool):
    global __enable
    __enable = v


def logEnable() -> bool:
    return __enable


def Debug(*args, **kwargs):
    if __enable:
        print(
            time.strftime(f"{BLUE}[DEBUG]"),
            *args,
            RESET,
            **kwargs
        )


def Info(*args, **kwargs):
    if __enable:
        print(
            time.strftime("[INFO] "),
            *args,
            **kwargs
        )


def Error(*args, **kwargs):
    if __enable:
        print(
            time.strftime(f"{RED}[ERROR]"),
            *args,
            RESET,
            **kwargs
        )



from ..type import Array
from ..util import UtString
from .amotion import AMotion
from .motion import Motion


class Live2DMotion(AMotion):
    MTN_PREFIX_VISIBLE = "VISIBLE:"
    MTN_PREFIX_LAYOUT = "LAYOUT:"
    MTN_PREFIX_FADE_IN = "FADEIN:"
    MTN_PREFIX_FADE_OUT = "FADEOUT:"

    def __init__(self):

        super().__init__()
        self.motions = Array()
        self.srcFps = 30
        self.maxLength = 0
        self.loop = False
        self.loopFadeIn = True
        self.loopDurationMSec = -1
        self.lastWeight = 0

    def getDurationMSec(self):
        return -1 if self.loop else self.loopDurationMSec

    def getLoopDurationMSec(self):
        return self.loopDurationMSec

    def updateParamExe(self, aJ, aN, aQ, a3):
        aO = aN - a3.startTimeMSec
        a0 = aO * self.srcFps / 1000
        aK = int(a0)
        aR = a0 - aK
        for aZ in range(0, len(self.motions), 1):
            aV = self.motions[aZ]
            aL = len(aV.values)
            aT = aV.paramIdStr
            if aV.mtnType == Motion.MOTION_TYPE_PARTS_VISIBLE:
                aX = aV.values[(aL - 1 if aK >= aL else aK)]
                aJ.setParamFloat(aT, aX)
            else:
                if Motion.MOTION_TYPE_LAYOUT_X <= aV.mtnType <= Motion.MOTION_TYPE_LAYOUT_SCALE_Y:
                    pass
                else:
                    aH = aJ.getParamIndex(aT)
                    a4 = aJ.getModelContext()
                    aY = a4.getParamMax(aH)
                    aW = a4.getParamMin(aH)
                    aM = 0.4
                    aS = aM * (aY - aW)
                    aU = a4.getParamFloat(aH)
                    a2 = aV.values[(aL - 1 if aK >= aL else aK)]
                    a1 = aV.values[(aL - 1 if aK + 1 >= aL else aK + 1)]
                    if (a2 < a1 and a1 - a2 > aS) or (a2 > a1 and a2 - a1 > aS):
                        aI = a2
                    else:
                        aI = a2 + (a1 - a2) * aR

                    aP = aU + (aI - aU) * aQ
                    aJ.setParamFloat(aT, aP)

        if aK >= self.maxLength:
            if self.loop:
                a3.startTimeMSec = aN
                if self.loopFadeIn:
                    a3.fadeInStartTimeMSec = aN
            else:
                a3.finished = True

        self.lastWeight = aQ

    def isLoop(self):
        return self.loop

    def setLoop(self, aH):
        self.loop = aH

    def isLoopFadeIn(self):
        return self.loopFadeIn

    def setLoopFadeIn(self, value):
        self.loopFadeIn = value

    @staticmethod
    def loadMotion(aT: bytes):
        mtn = Live2DMotion()
        aI = [0]
        aQ = len(aT)
        mtn.maxLength = 0
        aJ = 0
        while aJ < aQ:
            aL = aT[aJ]
            aS = chr(aL)
            if aS == "\n" or aS == "\r":
                aJ += 1
                continue

            if aS == "#":
                while aJ < aQ:
                    if chr(aT[aJ]) == "\n" or chr(aT[aJ]) == "\r":
                        break
                    aJ += 1

                aJ += 1
                continue

            if aS == "":
                aV = aJ
                aK = -1
                while aJ < aQ:
                    aS = chr(aT[aJ])
                    if aS == "\r" or aS == "\n":
                        break

                    if aS == "=":
                        aK = aJ
                        break
                    aJ += 1

                aP = False
                if aK >= 0:
                    if aK == aV + 4 and chr(aT[aV + 1]) == "f" and chr(aT[aV + 2]) == "p" and chr(aT[aV + 3]) == "s":
                        aP = True

                    aJ = aK + 1
                    while aJ < aQ:
                        aS = chr(aT[aJ])
                        if aS == "\r" or aS == "\n":
                            break

                        if aS == "," or aS == " " or aS == "\t":
                            aJ += 1
                            continue

                        aM = UtString.strToFloat(aT, aQ, aJ, aI)
                        if aI[0] > 0:
                            if aP and 5 < aM < 121:
                                mtn.srcFps = aM

                        aJ = aI[0]
                        aJ += 1

                while aJ < aQ:
                    if chr(aT[aJ]) == "\n" or chr(aT[aJ]) == "\r":
                        break
                    aJ += 1

                aJ += 1
                continue

            if (97 <= aL <= 122) or (65 <= aL <= 90) or aS == "_":
                aV = aJ
                aK = -1
                while aJ < aQ:
                    aS = chr(aT[aJ])
                    if aS == "\r" or aS == "\n":
                        break

                    if aS == "=":
                        aK = aJ
                        break
                    aJ += 1

                if aK >= 0:
                    aO = Motion()
                    if UtString.startswith(aT, aV, Live2DMotion.MTN_PREFIX_VISIBLE):
                        aO.mtnType = Motion.MOTION_TYPE_PARTS_VISIBLE
                        aO.paramIdStr = UtString.createString(aT, aV, aK - aV)
                    else:
                        if UtString.startswith(aT, aV, Live2DMotion.MTN_PREFIX_LAYOUT):
                            aO.paramIdStr = UtString.createString(aT, aV + 7, aK - aV - 7)
                            if UtString.startswith(aT, aV + 7, "ANCHOR_X"):
                                aO.mtnType = Motion.MOTION_TYPE_LAYOUT_ANCHOR_X
                            else:
                                if UtString.startswith(aT, aV + 7, "ANCHOR_Y"):
                                    aO.mtnType = Motion.MOTION_TYPE_LAYOUT_ANCHOR_Y
                                else:
                                    if UtString.startswith(aT, aV + 7, "SCALE_X"):
                                        aO.mtnType = Motion.MOTION_TYPE_LAYOUT_SCALE_X
                                    else:
                                        if UtString.startswith(aT, aV + 7, "SCALE_Y"):
                                            aO.mtnType = Motion.MOTION_TYPE_LAYOUT_SCALE_Y
                                        else:
                                            if UtString.startswith(aT, aV + 7, "AffineEnt"):
                                                aO.mtnType = Motion.MOTION_TYPE_LAYOUT_X
                                            else:
                                                if UtString.startswith(aT, aV + 7, "Y"):
                                                    aO.mtnType = Motion.MOTION_TYPE_LAYOUT_Y
                        else:
                            aO.mtnType = Motion.MOTION_TYPE_PARAM
                            aO.paramIdStr = UtString.createString(aT, aV, aK - aV)

                    mtn.motions.append(aO)
                    aU = 0
                    aR = []
                    aJ = aK + 1
                    while aJ < aQ:
                        aS = chr(aT[aJ])
                        if aS == "\r" or aS == "\n":
                            break

                        if aS == "," or aS == " " or aS == "\t":
                            aJ += 1
                            continue

                        aM = UtString.strToFloat(aT, aQ, aJ, aI)
                        if aI[0] > 0:
                            aR.append(aM)
                            aU += 1
                            aH = aI[0]
                            if aH < aJ:
                                print("invalid state during loadMotion\n")
                                break

                            aJ = aH - 1
                        aJ += 1

                    aO.values = aR
                    if aU > mtn.maxLength:
                        mtn.maxLength = aU

            aJ += 1

        mtn.loopDurationMSec = int((1000 * mtn.maxLength) / mtn.srcFps)
        return mtn



from abc import ABC, abstractmethod

from ..util import UtSystem
from ..util.ut_motion import UtMotion


class AMotion(ABC):

    def __init__(self):
        self.fadeInMSec = 1000
        self.fadeOutMSec = 1000
        self.weight = 1

    def setFadeIn(self, aH):
        self.fadeInMSec = aH

    def setFadeOut(self, aH):
        self.fadeOutMSec = aH

    def setWeight(self, aH):
        self.weight = aH

    def getFadeOut(self):
        return self.fadeOutMSec

    def getWeight(self):
        return self.weight

    def getDurationMSec(self):
        return -1

    def getLoopDurationMSec(self):
        return -1

    def updateParam(self, aJ, aN):
        if not aN.available or aN.finished:
            return

        aL = UtSystem.getUserTimeMSec()
        if aN.startTimeMSec < 0:
            aN.startTimeMSec = aL
            aN.fadeInStartTimeMSec = aL
            aM = self.getDurationMSec()
            if aN.endTimeMSec < 0:
                aN.endTimeMSec = -1 if (aM <= 0) else aN.startTimeMSec + aM

        aI = self.weight
        aH = 1 if (self.fadeInMSec == 0) else UtMotion.getEasingSine(((aL - aN.fadeInStartTimeMSec) / self.fadeInMSec))
        aK = 1 if (self.fadeOutMSec == 0 or aN.endTimeMSec < 0) else UtMotion.getEasingSine(((aN.endTimeMSec - aL) / self.fadeOutMSec))
        aI = aI * aH * aK
        if not (0 <= aI <= 1):
            print("### assert!! ### ")

        self.updateParamExe(aJ, aL, aI, aN)
        if 0 < aN.endTimeMSec < aL:
            aN.finished = True

    @abstractmethod
    def updateParamExe(self, aH, aI, aJ, aK):
        pass

    @staticmethod
    def getEasing(t, totalTime, accelerateTime):
        aQ = t / totalTime
        a1 = accelerateTime / totalTime
        aU = a1
        aZ = 1 / 3
        aR = 2 / 3
        a0 = 1 - (1 - a1) * (1 - a1)
        a2 = 1 - (1 - aU) * (1 - aU)
        aM = 0
        aL = ((1 - a1) * aZ) * a0 + (aU * aR + (1 - aU) * aZ) * (1 - a0)
        aK = (aU + (1 - aU) * aR) * a2 + (a1 * aZ + (1 - a1) * aR) * (1 - a2)
        aJ = 1
        aY = aJ - 3 * aK + 3 * aL - aM
        aX = 3 * aK - 6 * aL + 3 * aM
        aW = 3 * aL - 3 * aM
        aV = aM
        if aQ <= 0:
            return 0
        elif aQ >= 1:
            return 1

        aS = aQ
        aI = aS * aS
        aH = aS * aI
        aT = aY * aH + aX * aI + aW * aS + aV
        return aT



from .params import *  # pyinstaller may not find it (hidden import)
from .live2d import *


class MotionPriority:
    NONE = 0
    IDLE = 1
    NORMAL = 2
    FORCE = 3


class MotionGroup:
    IDLE = "Idle"
    TAP_HEAD = "TapHead"


class HitArea:
    HEAD = MotionGroup.TAP_HEAD


LIVE2D_VERSION = 3



from framework.matrix.l2d_matrix44 import L2DMatrix44


class L2DViewMatrix(L2DMatrix44):

    def __init__(self):
        super().__init__()
        self.screenLeft = None
        self.screenRight = None
        self.screenTop = None
        self.screenBottom = None
        self.maxLeft = None
        self.maxRight = None
        self.maxTop = None
        self.maxBottom = None
        self.max = float('inf')
        self.min = 0

    def getMaxScale(self):
        return self.max

    def getMinScale(self):
        return self.min

    def setMaxScale(self, v):
        self.max = v

    def setMinScale(self, v):
        self.min = v

    def isMaxScale(self):
        return self.getScaleX() == self.max

    def isMinScale(self):
        return self.getScaleX() == self.min

    def adjustTranslate(self, shift_x, shift_y):
        if self.tr[0] * self.maxLeft + (self.tr[12] + shift_x) > self.screenLeft:
            shift_x = self.screenLeft - self.tr[0] * self.maxLeft - self.tr[12]
        if self.tr[0] * self.maxRight + (self.tr[12] + shift_x) < self.screenRight:
            shift_x = self.screenRight - self.tr[0] * self.maxRight - self.tr[12]
        if self.tr[5] * self.maxTop + (self.tr[13] + shift_y) < self.screenTop:
            shift_y = self.screenTop - self.tr[5] * self.maxTop - self.tr[13]
        if self.tr[5] * self.maxBottom + (self.tr[13] + shift_y) > self.screenBottom:
            shift_y = self.screenBottom - self.tr[5] * self.maxBottom - self.tr[13]
        tr1 = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, shift_x, shift_y, 0, 1]
        L2DMatrix44.mul(tr1, self.tr, self.tr)

    def adjustScale(self, cx, cy, scale):
        target_scale = scale * self.tr[0]
        if target_scale < self.min:
            if self.tr[0] > 0:
                scale = self.min / self.tr[0]
        elif target_scale > self.max:
            if self.tr[0] > 0:
                scale = self.max / self.tr[0]

        tr1 = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, cx, cy, 0, 1]
        tr2 = [scale, 0, 0, 0, 0, scale, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]
        tr3 = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, -cx, -cy, 0, 1]
        L2DMatrix44.mul(tr3, self.tr, self.tr)
        L2DMatrix44.mul(tr2, self.tr, self.tr)
        L2DMatrix44.mul(tr1, self.tr, self.tr)

    def setScreenRect(self, left, right, bottom, top):
        self.screenLeft = left
        self.screenRight = right
        self.screenTop = top
        self.screenBottom = bottom

    def setMaxScreenRect(self, left, right, bottom, top):
        self.maxLeft = left
        self.maxRight = right
        self.maxTop = top
        self.maxBottom = bottom

    def getScreenLeft(self):
        return self.screenLeft

    def getScreenRight(self):
        return self.screenRight

    def getScreenBottom(self):
        return self.screenBottom

    def getScreenTop(self):
        return self.screenTop

    def getMaxLeft(self):
        return self.maxLeft

    def getMaxRight(self):
        return self.maxRight

    def getMaxBottom(self):
        return self.maxBottom

    def getMaxTop(self):
        return self.maxTop



class Motion:
    MOTION_TYPE_PARAM = 0
    MOTION_TYPE_PARTS_VISIBLE = 1
    MOTION_TYPE_LAYOUT_X = 100
    MOTION_TYPE_LAYOUT_Y = 101
    MOTION_TYPE_LAYOUT_ANCHOR_X = 102
    MOTION_TYPE_LAYOUT_ANCHOR_Y = 103
    MOTION_TYPE_LAYOUT_SCALE_X = 104
    MOTION_TYPE_LAYOUT_SCALE_Y = 105

    def __init__(self):
        self.paramIdStr = None
        self.values = None
        self.mtnType = None



from .log import __ut_log as log
from .ut_interpolate import UtInterpolate
from .ut_math import UtMath
from .ut_string import UtString
from .ut_system import UtSystem

__all__ = ['UtMath', 'UtString', 'UtInterpolate', 'UtSystem', 'log']



from typing import Tuple, TYPE_CHECKING

from .framework import L2DMatrix44

if TYPE_CHECKING:
    from .framework.matrix import L2DModelMatrix



class MatrixManager:

    def __init__(self):
        self.__projection = L2DMatrix44()
        self.__screenToScene = L2DMatrix44()
        self.__ww = 600
        self.__wh = 600
        self.__offsetX = 0
        self.__offsetY = 0
        self.__scale = 1

    def getWidth(self):
        return self.__ww

    def getHeight(self):
        return self.__wh

    def onResize(self, width: int, height: int):
        self.__ww = width
        self.__wh = height

        ratio = float(width) / float(height)
        left = -ratio
        right = ratio
        bottom = -1.0
        top = 1.0

        self.__screenToScene.identity()
        self.__screenToScene.multTranslate(-width / 2, -height / 2)

        if width > height:
            sw = abs(right - left)
            self.__screenToScene.multScale(sw / width, -sw / width)
        else:
            sh = abs(top - bottom)
            self.__screenToScene.multScale(sh / height, -sh / height)

    def screenToScene(self, scr_x: float, scr_y: float) -> Tuple[float, float]:
        return self.__screenToScene.transformX(scr_x), self.__screenToScene.transformY(scr_y)

    def invertTransform(self, src_x, src_y) -> Tuple[float, float]:
        return self.__projection.invertTransformX(src_x), self.__projection.invertTransformY(src_y)

    def setScale(self, scale: float):
        self.__scale = scale

    def setOffset(self, dx: float, dy: float):
        self.__offsetX = dx
        self.__offsetY = dy

    def getMvp(self, model_matrix: 'L2DModelMatrix') -> list:
        self.__projection.identity()

        if self.__wh > self.__ww:
            model_matrix.setWidth(2.0)
            self.__projection.multScale(1.0, self.__ww / self.__wh)
        else:
            self.__projection.multScale(self.__wh / self.__ww, 1.0)

        self.__projection.multScale(self.__scale, self.__scale)
        self.__projection.translate(self.__offsetX, self.__offsetY)

        self.__projection.mul(self.__projection.getArray(), model_matrix.getArray(), self.__projection.getArray())
        return self.__projection.getArray()



import OpenGL.GL as gl
import numpy as np


class Live2DGLWrapper:
    """
    live2d 用到的所有 api
    """
    FRAMEBUFFER = gl.GL_FRAMEBUFFER
    RENDERBUFFER = gl.GL_RENDERBUFFER
    COLOR_BUFFER_BIT = gl.GL_COLOR_BUFFER_BIT
    RGBA4 = gl.GL_RGBA4
    COLOR_ATTACHMENT0 = gl.GL_COLOR_ATTACHMENT0
    RGBA = gl.GL_RGBA
    UNSIGNED_BYTE = gl.GL_UNSIGNED_BYTE
    TEXTURE_2D = gl.GL_TEXTURE_2D
    TEXTURE_MIN_FILTER = gl.GL_TEXTURE_MIN_FILTER
    TEXTURE_MAG_FILTER = gl.GL_TEXTURE_MAG_FILTER
    LINEAR = gl.GL_LINEAR
    CLAMP_TO_EDGE = gl.GL_CLAMP_TO_EDGE
    TEXTURE_WRAP_S = gl.GL_TEXTURE_WRAP_S
    TEXTURE_WRAP_T = gl.GL_TEXTURE_WRAP_T
    VERTEX_SHADER = gl.GL_VERTEX_SHADER
    FRAGMENT_SHADER = gl.GL_FRAGMENT_SHADER
    COMPILE_STATUS = gl.GL_COMPILE_STATUS
    LINK_STATUS = gl.GL_LINK_STATUS
    SCISSOR_TEST = gl.GL_SCISSOR_TEST
    STENCIL_TEST = gl.GL_STENCIL_TEST
    DEPTH_TEST = gl.GL_DEPTH_TEST
    CW = gl.GL_CW
    CCW = gl.GL_CCW
    BLEND = gl.GL_BLEND
    ARRAY_BUFFER = gl.GL_ARRAY_BUFFER
    ELEMENT_ARRAY_BUFFER = gl.GL_ELEMENT_ARRAY_BUFFER
    DYNAMIC_DRAW = gl.GL_DYNAMIC_DRAW
    FLOAT = gl.GL_FLOAT
    TEXTURE1 = gl.GL_TEXTURE1
    TEXTURE2 = gl.GL_TEXTURE2
    CULL_FACE = gl.GL_CULL_FACE
    ONE = gl.GL_ONE
    SRC_ALPHA = gl.GL_SRC_ALPHA
    ONE_MINUS_SRC_ALPHA = gl.GL_ONE_MINUS_SRC_ALPHA
    DST_COLOR = gl.GL_DST_COLOR
    ZERO = gl.GL_ZERO
    FUNC_ADD = gl.GL_FUNC_ADD
    TRIANGLES = gl.GL_TRIANGLES
    UNSIGNED_SHORT = gl.GL_UNSIGNED_SHORT
    FRAMEBUFFER_BINDING = gl.GL_FRAMEBUFFER_BINDING
    DEPTH_BUFFER_BIT = gl.GL_DEPTH_BUFFER_BIT
    ONE_MINUS_SRC_COLOR = gl.GL_ONE_MINUS_SRC_COLOR

    def __init__(self):
        self.width = 0
        self.height = 0

    def resize(self, w, h):
        self.width = w
        self.height = h

    @staticmethod
    def getAttribLocation(program, name):
        return gl.glGetAttribLocation(program, name)

    @staticmethod
    def getUniformLocation(program, name):
        return gl.glGetUniformLocation(program, name)

    @staticmethod
    def createFramebuffer():
        return gl.glGenFramebuffers(1)

    @staticmethod
    def bindFramebuffer(t, fbo):
        gl.glBindFramebuffer(t, fbo)

    @staticmethod
    def createRenderbuffer():
        return gl.glGenRenderbuffers(1)

    @staticmethod
    def bindRenderbuffer(t, rbo):
        gl.glBindRenderbuffer(t, rbo)

    @staticmethod
    def renderbufferStorage(t, fat, width, height):
        gl.glRenderbufferStorage(t, fat, width, height)

    @staticmethod
    def framebufferRenderbuffer(t, attachment, rbt, rb):
        gl.glFramebufferRenderbuffer(t, attachment, rbt, rb)

    @staticmethod
    def createTexture():
        return gl.glGenTextures(1)

    @staticmethod
    def bindTexture(t, tid):
        gl.glBindTexture(t, tid)

    @staticmethod
    def texImage2D(*args):
        gl.glTexImage2D(*args)

    @staticmethod
    def texParameteri(*args):
        gl.glTexParameteri(*args)

    @staticmethod
    def framebufferTexture2D(*args):
        gl.glFramebufferTexture2D(*args)

    @staticmethod
    def createProgram():
        return gl.glCreateProgram()

    @staticmethod
    def compileShader(*args):
        gl.glCompileShader(*args)

    @staticmethod
    def createShader(*args):
        return gl.glCreateShader(*args)

    @staticmethod
    def shaderSource(*args):
        gl.glShaderSource(*args)

    @staticmethod
    def getShaderParameter(*args):
        return gl.glGetShaderiv(*args)

    @staticmethod
    def getShaderInfoLog(*args):
        return gl.glGetShaderInfoLog(*args)

    @staticmethod
    def attachShader(*args):
        return gl.glAttachShader(*args)

    @staticmethod
    def linkProgram(*args):
        gl.glLinkProgram(*args)

    @staticmethod
    def getProgramParameter(*args):
        return gl.glGetProgramiv(*args)

    @staticmethod
    def getProgramInfoLog(*args):
        return gl.glGetProgramInfoLog(*args)

    @staticmethod
    def disable(t):
        gl.glDisable(t)

    @staticmethod
    def bindBuffer(*args):
        gl.glBindBuffer(*args)

    @staticmethod
    def enable(t):
        gl.glEnable(t)

    @staticmethod
    def colorMask(*args):
        gl.glColorMask(*args)

    @staticmethod
    def frontFace(t):
        gl.glFrontFace(t)

    @staticmethod
    def useProgram(p):
        gl.glUseProgram(p)

    @staticmethod
    def createBuffer():
        return gl.glGenBuffers(1)

    @staticmethod
    def bufferData(t, data, usage):
        if t == Live2DGLWrapper.ARRAY_BUFFER:
            buf = np.array(data, dtype=np.float32)
        elif t == Live2DGLWrapper.ELEMENT_ARRAY_BUFFER:
            buf = np.array(data, dtype=np.uint16)
        else:
            raise Exception()

        gl.glBufferData(t, buf.nbytes, buf, usage)

    @staticmethod
    def enableVertexAttribArray(vao):
        gl.glEnableVertexAttribArray(vao)

    @staticmethod
    def vertexAttribPointer(*args):
        gl.glVertexAttribPointer(*args)

    @staticmethod
    def activeTexture(t):
        gl.glActiveTexture(t)

    @staticmethod
    def uniform1i(*args):
        gl.glUniform1i(*args)

    @staticmethod
    def uniformMatrix4fv(loc, transpose, value):
        buf = np.array(value, dtype=np.float32)
        gl.glUniformMatrix4fv(loc, 1, transpose, buf)

    @staticmethod
    def uniform4f(loc, *args):
        gl.glUniform4f(loc, *args)

    @staticmethod
    def blendEquationSeparate(a, b):
        gl.glBlendEquationSeparate(a, b)

    @staticmethod
    def blendFuncSeparate(a, b, c, d):
        gl.glBlendFuncSeparate(a, b, c, d)

    @staticmethod
    def drawElements(t, size, dt, data):
        gl.glDrawElements(t, size, dt, data)

    @staticmethod
    def getParameter(t):
        return gl.glGetIntegerv(t)

    @staticmethod
    def viewport(a, b, c, d):
        gl.glViewport(a, b, c, d)

    @staticmethod
    def clearColor(a, b, c, d):
        gl.glClearColor(a, b, c, d)

    @staticmethod
    def clear(t):
        gl.glClear(t)

    @staticmethod
    def deleteFramebuffer(t):
        gl.glDeleteFramebuffers(1, t)

    @staticmethod
    def deleteShader(s):
        gl.glDeleteShader(s)

    @staticmethod
    def deleteTexture(t):
        gl.glDeleteTextures(1, t)

    @staticmethod
    def deleteBuffer(b):
        gl.glDeleteBuffers(1, b)

    @staticmethod
    def deleteProgram(p):
        gl.glDeleteProgram(p)

    @staticmethod
    def deleteRenderbuffer(r):
        gl.glDeleteRenderbuffers(1, r)



import math

from ..type import Array
from ..util import UtMath
from .physics_constants import *
from .physics_point import PhysicsPoint
from .physics_src import PhysicsSrc
from .physics_target import PhysicsTarget


class PhysicsHair:
    SRC_TO_X = SRC_TO_X
    SRC_TO_Y = SRC_TO_Y
    SRC_TO_G_ANGLE = SRC_TO_G_ANGLE

    TARGET_FROM_ANGLE = TARGET_FROM_ANGLE
    TARGET_FROM_ANGLE_V = TARGET_FROM_ANGLE_V

    def __init__(self):
        self.p1 = PhysicsPoint()
        self.p2 = PhysicsPoint()
        self.Fo_ = 0
        self.Db_ = 0
        self.L2_ = 0
        self.M2_ = 0
        self.ks_ = 0
        self._9b = 0
        self.iP_ = 0
        self.iT_ = 0
        self.lL_ = Array()
        self.qP_ = Array()
        self.setup(0.3, 0.5, 0.1)

    def setup(self, aJ=None, aI=None, aH=None):
        self.ks_ = self.Yb_()
        self.p2.setupLast()
        if aH is not None:
            self.Fo_ = aJ
            self.L2_ = aI
            self.p1.mass = aH
            self.p2.mass = aH
            self.p2.y = aJ
            self.setup()

    def getPhysicsPoint1(self):
        return self.p1

    def getPhysicsPoint2(self):
        return self.p2

    def qr_(self):
        return self.Db_

    def pr_(self, aH):
        self.Db_ = aH

    def _5r(self):
        return self.M2_

    def Cs_(self):
        return self._9b

    def Yb_(self):
        return -180 * (math.atan2(self.p1.x - self.p2.x, -(self.p1.y - self.p2.y))) / math.pi

    def addSrcParam(self, aJ, aH, aL, aI):
        aK = PhysicsSrc(aJ, aH, aL, aI)
        self.lL_.append(aK)

    def addTargetParam(self, aJ, aH, aK, aI):
        aL = PhysicsTarget(aJ, aH, aK, aI)
        self.qP_.append(aL)

    def update(self, aI, aL):
        if self.iP_ == 0:
            self.iP_ = self.iT_ = aL
            self.Fo_ = (math.sqrt((self.p1.x - self.p2.x) * (self.p1.x - self.p2.x) + (self.p1.y - self.p2.y) * (
                    self.p1.y - self.p2.y)))
            return

        aK = (aL - self.iT_) / 1000
        if aK != 0:
            for aJ in range(len(self.lL_) - 1, 0 - 1, -1):
                aM = self.lL_[aJ]
                aM.update(aI, self)

            self.oo_(aI, aK)
            self.M2_ = self.Yb_()
            self._9b = (self.M2_ - self.ks_) / aK
            self.ks_ = self.M2_

        for aJ in range(len(self.qP_) - 1, 0 - 1, -1):
            aH = self.qP_[aJ]
            aH.update(aI, self)

        self.iT_ = aL

    def oo_(self, aN, aI):
        if aI < 0.033:
            aI = 0.033

        aU = 1 / aI
        self.p1.vx = (self.p1.x - self.p1.lastX) * aU
        self.p1.vy = (self.p1.y - self.p1.lastY) * aU
        self.p1.ax = (self.p1.vx - self.p1.lastVX) * aU
        self.p1.ay = (self.p1.vy - self.p1.lastVY) * aU
        self.p1.fx = self.p1.ax * self.p1.mass
        self.p1.fy = self.p1.ay * self.p1.mass
        self.p1.setupLast()
        aM = -(math.atan2((self.p1.y - self.p2.y), self.p1.x - self.p2.x))
        aR = math.cos(aM)
        aH = math.sin(aM)
        aW = 9.8 * self.p2.mass
        aQ = (self.Db_ * UtMath.DEG_TO_RAD)
        aP = (aW * math.cos(aM - aQ))
        aL = (aP * aH)
        aV = (aP * aR)
        aK = (-self.p1.fx * aH * aH)
        aT = (-self.p1.fy * aH * aR)
        aJ = (-self.p2.vx * self.L2_)
        aS = (-self.p2.vy * self.L2_)
        self.p2.fx = (aL + aK + aJ)
        self.p2.fy = (aV + aT + aS)
        self.p2.ax = self.p2.fx / self.p2.mass
        self.p2.ay = self.p2.fy / self.p2.mass
        self.p2.vx += self.p2.ax * aI
        self.p2.vy += self.p2.ay * aI
        self.p2.x += self.p2.vx * aI
        self.p2.y += self.p2.vy * aI
        aO = (math.sqrt(
            (self.p1.x - self.p2.x) * (self.p1.x - self.p2.x) + (self.p1.y - self.p2.y) * (self.p1.y - self.p2.y)))
        self.p2.x = self.p1.x + self.Fo_ * (self.p2.x - self.p1.x) / aO
        self.p2.y = self.p1.y + self.Fo_ * (self.p2.y - self.p1.y) / aO
        self.p2.vx = (self.p2.x - self.p2.lastX) * aU
        self.p2.vy = (self.p2.y - self.p2.lastY) * aU
        self.p2.setupLast()



from .alive2d_model import ALive2DModel
from .live2d_gl_wrapper import Live2DGLWrapper
from .live2d_model_opengl import Live2DModelOpenGL
from .live2d import Live2D
from .util import log, UtSystem
from .motion import Live2DMotion, AMotion, MotionQueueManager
from .physics import PhysicsHair
from .id import Id
from .type import Array



from .l2d_matrix44 import L2DMatrix44


class L2DModelMatrix(L2DMatrix44):

    def __init__(self, w, h):
        super().__init__()
        self.width = w
        self.height = h
        self.ocx = 0
        self.ocy = 0

    def setPosition(self, x, y):
        self.translate(x, y)

    def setCenterPosition(self, x, y):
        self.ocx = x
        self.ocy = y
        w = self.width * self.getScaleX()
        h = self.height * self.getScaleY()
        self.translate(x - w / 2, y - h / 2)

    def top(self, y):
        self.setY(y)

    def bottom(self, y):
        h = self.height * self.getScaleY()
        self.translateY(y - h)

    def left(self, x):
        self.setX(x)

    def right(self, x):
        w = self.width * self.getScaleX()
        self.translateX(x - w)

    def centerX(self, x):
        w = self.width * self.getScaleX()
        self.translateX(x - w / 2)

    def centerY(self, y):
        h = self.height * self.getScaleY()
        self.translateY(y - h / 2)

    def setX(self, x):
        self.translateX(x)

    def setY(self, y):
        self.translateY(y)

    def setHeight(self, h):
        scale_x = h / self.height
        scale_y = -scale_x
        self.scale(scale_x, scale_y)

    def setWidth(self, w):
        scale_x = w / self.width
        scale_y = -scale_x
        self.scale(scale_x, scale_y)



class ClipRectF:

    def __init__(self):
        self.x = None
        self.y = None
        self.width = None
        self.height = None

    def getRight(self):
        return self.x + self.width

    def getBottom(self):
        return self.y + self.height

    def expand(self, aH, aI):
        self.x -= aH
        self.y -= aI
        self.width += aH * 2
        self.height += aI * 2

    def setRect(self, other):
        self.x = other.x
        self.y = other.y
        self.width = other.width
        self.height = other.height



from abc import ABC, abstractmethod

class IPhysicsParam(ABC):

    def __init__(self, paramId, scale, weight):
        self.paramId = paramId
        self.scale = scale
        self.weight = weight

    @abstractmethod
    def update(self, aI, aH):
        pass



from typing import TYPE_CHECKING

from .deformer import Deformer
from .deformer_context import DeformerContext

if TYPE_CHECKING:
    from .roation_deformer import AffineEnt


class RotationContext(DeformerContext):

    def __init__(self, dfm):
        super().__init__(dfm)

        self.tmpDeformerIndex = Deformer.DEFORMER_INDEX_NOT_INIT
        self.interpolatedAffine: None | AffineEnt = None
        self.transformedAffine: None | AffineEnt = None



from typing import Union
import json
import math

from .framework import Live2DFramework


class ModelSettingJson():
    
    def __init__(self):
        self.NAME = "name"
        self.ID = "id"
        self.MODEL = "model"
        self.TEXTURES = "textures"
        self.HIT_AREAS = "hit_areas"
        self.PHYSICS = "physics"
        self.POSE = "pose"
        self.EXPRESSIONS = "expressions"
        self.MOTION_GROUPS = "motions"
        self.SOUND = "sound"
        self.FADE_IN = "fade_in"
        self.FADE_OUT = "fade_out"
        self.LAYOUT = "layout"
        self.INIT_PARAM = "init_param"
        self.INIT_PARTS_VISIBLE = "init_parts_visible"
        self.VALUE = "val"
        self.FILE = "file"
        self.json = {}
    
    
    def loadModelSetting(self, path) -> None:
        pm = Live2DFramework.getPlatformManager()
        data = pm.loadBytes(path)
        self.json = json.loads(data)
    
    
    def getTextureFile(self, n) -> Union[str, None]:
        if self.json.get(self.TEXTURES) is None or self.json[self.TEXTURES][n] is None:
            return None
        return self.json[self.TEXTURES][n]
    
    
    def getModelFile(self):
        return self.json[self.MODEL]
    
    
    def getTextureNum(self) -> int:
        if self.json.get(self.TEXTURES) is None:
            return 0
        return len(self.json[self.TEXTURES])
    
    
    def getHitAreaNum(self):
        if self.json.get(self.HIT_AREAS, None) is None:
            return 0
        return len(self.json[self.HIT_AREAS])
    
    
    def getHitAreaID(self, n):
        if self.json.get(self.HIT_AREAS, None) is None or self.json[self.HIT_AREAS].get(n, None) is None:
            return None
        return self.json[self.HIT_AREAS][n][self.ID]
    
    
    def getHitAreaName(self, n):
        if self.json.get(self.HIT_AREAS, None) is None or self.json[self.HIT_AREAS].get(n, None) is None:
            return None
        return self.json[self.HIT_AREAS][n][self.NAME]
    
    
    def getPhysicsFile(self):
        return self.json.get(self.PHYSICS)
    
    
    def getPoseFile(self):
        return self.json.get(self.POSE)
    
    
    def getExpressionNum(self):
        return 0 if (self.json.get(self.EXPRESSIONS, None) is None) else len(self.json[self.EXPRESSIONS])
    
    
    def getExpressionFile(self, n):
        if self.json.get(self.EXPRESSIONS, None) is None:
            return None
        return self.json[self.EXPRESSIONS][n][self.FILE]
    
    
    def getExpressionName(self, n):
        if self.json.get(self.EXPRESSIONS, None) is None:
            return None
        return self.json[self.EXPRESSIONS][n][self.NAME]
    
    
    def getLayout(self):
        return self.json.get(self.LAYOUT)
    
    
    def getInitParamNum(self):
        return 0 if (self.json.get(self.INIT_PARAM, None) is None) else len(self.json[self.INIT_PARAM])
    
    
    def getMotionNum(self, name):
        if self.json.get(self.MOTION_GROUPS, None) is None or self.json[self.MOTION_GROUPS].get(name, None) is None:
            return 0
        return len(self.json[self.MOTION_GROUPS][name])
    
    
    def getMotionFile(self, name, n):
        if self.json.get(self.MOTION_GROUPS, None) is None or self.json[self.MOTION_GROUPS].get(name, None) is None or self.json[self.MOTION_GROUPS][name][n] is None:
            return None
        return self.json[self.MOTION_GROUPS][name][n][self.FILE]
    
    
    def getMotionSound(self, name, n):
        if self.json.get(self.MOTION_GROUPS, None) is None or self.json[self.MOTION_GROUPS].get(name, None) is None or self.json[self.MOTION_GROUPS][name][n] is None or self.json[self.MOTION_GROUPS][name][n].get(self.SOUND, None) is None:
            return None
        return self.json[self.MOTION_GROUPS][name][n][self.SOUND]
    
    
    def getMotionFadeIn(self, name, n):
        if self.json.get(self.MOTION_GROUPS, None) is None or self.json[self.MOTION_GROUPS].get(name, None) is None or self.json[self.MOTION_GROUPS][name][n] is None or self.json[self.MOTION_GROUPS][name][n].get(self.FADE_IN, None) is None:
            return 1000
        return self.json[self.MOTION_GROUPS][name][n][self.FADE_IN]
    
    
    def getMotionFadeOut(self, name, n):
        if self.json.get(self.MOTION_GROUPS, None) is None or self.json[self.MOTION_GROUPS].get(name, None) is None or self.json[self.MOTION_GROUPS][name][n] is None or self.json[self.MOTION_GROUPS][name][n].get(self.FADE_OUT, None) is None:
            return 1000
        return self.json[self.MOTION_GROUPS][name][n][self.FADE_OUT]


    def getMotionNames(self):
        if self.json.get(self.MOTION_GROUPS, None) is None:
            return None

        return tuple(self.json[self.MOTION_GROUPS].keys())
    
    
    def getInitParamID(self, n):
        if self.json.get(self.INIT_PARAM, None) is None or self.json[self.INIT_PARAM].get(n, None) is None:
            return None
        return self.json[self.INIT_PARAM][n][self.ID]
    
    
    def getInitParamValue(self, n):
        if self.json.get(self.INIT_PARAM, None) is None or self.json[self.INIT_PARAM].get(n, None) is None:
            return math.nan
        return self.json[self.INIT_PARAM][n][self.VALUE]
    
    
    def getInitPartsVisibleNum(self):
        return 0 if (self.json.get(self.INIT_PARTS_VISIBLE, None) is None) else len(self.json[self.INIT_PARTS_VISIBLE])
    
    
    def getInitPartsVisibleID(self, n):
        if self.json.get(self.INIT_PARTS_VISIBLE, None) is None or self.json[self.INIT_PARTS_VISIBLE].get(n, None) is None:
            return None
        return self.json[self.INIT_PARTS_VISIBLE][n][self.ID]
    
    def getInitPartsVisibleValue(self, n):
        if self.json.get(self.INIT_PARTS_VISIBLE, None) is None or self.json[self.INIT_PARTS_VISIBLE].get(n, None) is None:
            return math.nan
        return self.json[self.INIT_PARTS_VISIBLE][n][self.VALUE]




from abc import abstractmethod

from ..DEF import LIVE2D_FORMAT_VERSION_AVAILABLE
from ..id import Id
from ..io.iserializable import ISerializable
from ..live2d import Live2D
from ..util import UtInterpolate


class IDrawData(ISerializable):
    DEFORMER_INDEX_NOT_INIT = -2
    DEFAULT_ORDER = 500
    TYPE_MESH = 2
    totalMinOrder = DEFAULT_ORDER
    totalMaxOrder = DEFAULT_ORDER

    def __init__(self):
        super().__init__()
        self.clipIDList = None
        self.clipID = None
        self.id = None
        self.targetId = None
        self.pivotMgr = None
        self.averageDrawOrder = None
        self.pivotDrawOrders = None
        self.pivotOpacities = None

    def read(self, aH):
        self.id = aH.readObject()
        self.targetId = aH.readObject()
        self.pivotMgr = aH.readObject()
        self.averageDrawOrder = aH.readInt32()
        self.pivotDrawOrders = aH.readInt32Array()
        self.pivotOpacities = aH.readFloat32Array()
        if aH.getFormatVersion() >= LIVE2D_FORMAT_VERSION_AVAILABLE:
            self.clipID = aH.readObject()
            self.clipIDList = self.convertClipIDForV2_11(self.clipID)
        else:
            self.clipIDList = None

        IDrawData.setDrawOrders(self.pivotDrawOrders)

    def getClipIDList(self):
        return self.clipIDList

    @staticmethod
    def convertClipIDForV2_11(s):
        ls = []
        if s is None:
            return None

        if len(s.id) == 0:
            return None

        if not s.id.find(','):
            ls.append(s.id)
            return ls

        ls = s.id.split(",")
        return ls

    def setupInterpolate(self, aI, aH):
        aH.paramOutside = False
        aH.interpolatedDrawOrder = UtInterpolate.interpolateInt(aI, self.pivotMgr, aH.paramOutside,
                                                                self.pivotDrawOrders)
        if not Live2D.L2D_OUTSIDE_PARAM_AVAILABLE and aH.paramOutside:
            return

        aH.interpolatedOpacity = UtInterpolate.interpolateFloat(aI, self.pivotMgr, aH.paramOutside, self.pivotOpacities)

    @abstractmethod
    def setupTransform(self, mc, dc=None):
        pass

    def getId(self):
        return self.id

    def setId(self, value):
        self.id = value

    @staticmethod
    def getOpacity(ctx):
        return ctx.interpolatedOpacity

    @staticmethod
    def getDrawOrder(ctx):
        return ctx.interpolatedDrawOrder

    def getTargetId(self):
        return self.targetId

    def setTargetId(self, aH):
        self.targetId = aH

    def needTransform(self):
        return self.targetId is not None and (self.targetId != Id.DST_BASE_ID())

    @abstractmethod
    def getType(self):
        pass

    @staticmethod
    def setDrawOrders(orders):
        for i in range(len(orders) - 1, 0 - 1, -1):
            order = orders[i]
            if order < IDrawData.totalMinOrder:
                IDrawData.totalMinOrder = order
            else:
                if order > IDrawData.totalMaxOrder:
                    IDrawData.totalMaxOrder = order

    @staticmethod
    def getTotalMinOrder():
        return IDrawData.totalMinOrder

    @staticmethod
    def getTotalMaxOrder():
        return IDrawData.totalMaxOrder



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ..platform_manager import PlatformManager

class Live2DFramework:
    __platformManager: Optional['PlatformManager'] = None

    @staticmethod
    def getPlatformManager() -> Optional['PlatformManager']:
        return Live2DFramework.__platformManager

    @staticmethod
    def setPlatformManager(platformManager: Optional['PlatformManager']) -> None:
        Live2DFramework.__platformManager = platformManager



from ..io.iserializable import ISerializable


class Avatar(ISerializable):

    def __init__(self):
        self.id = None
        self.deformerList = None
        self.drawDataList = None

    def getDeformer(self):
        return self.deformerList

    def getDrawDataList(self):
        return self.drawDataList

    def read(self, br):
        self.id = br.readObject()
        self.drawDataList = br.readObject()
        self.deformerList = br.readObject()

    def replacePartsData(self, parts):
        parts.setDeformer(self.deformerList)
        parts.setDrawDataList(self.drawDataList)
        self.deformerList = None
        self.drawDataList = None



from ..deformer import RotationDeformer, AffineEnt, WarpDeformer
from ..draw import Mesh
from ..model import ModelImpl, Avatar, PartsData
from ..param import PivotManager, ParamPivots, ParamDefFloat, ParamDefSet


class Live2DObjectFactory:

    @staticmethod
    def create(clsNo):
        if clsNo < 100:
            if clsNo == 65:
                return WarpDeformer()
            elif clsNo == 66:
                return PivotManager()
            elif clsNo == 67:
                return ParamPivots()
            elif clsNo == 68:
                return RotationDeformer()
            elif clsNo == 69:
                return AffineEnt()
            elif clsNo == 70:
                return Mesh()
        elif clsNo < 150:
            if clsNo == 131:
                return ParamDefFloat()
            elif clsNo == 133:
                return PartsData()
            elif clsNo == 136:
                return ModelImpl()
            elif clsNo == 137:
                return ParamDefSet()
            elif clsNo == 142:
                return Avatar()

        raise RuntimeError("Unknown class ID: " + str(clsNo))



from typing import Optional, List, TYPE_CHECKING

from ..io.iserializable import ISerializable
from ..type import Array
from ..param import ParamDefSet

if TYPE_CHECKING:
    from .part import PartsData


class ModelImpl(ISerializable):

    def __init__(self):
        self.paramDefSet: Optional[ParamDefSet] = None
        self.partsDataList: Optional[List['PartsData']] = None
        self.canvasWidth: int = 400
        self.canvasHeight: int = 400

    def initDirect(self):
        if self.paramDefSet is None:
            self.paramDefSet = ParamDefSet()

        if self.partsDataList is None:
            self.partsDataList = Array()

    def getCanvasWidth(self):
        return self.canvasWidth

    def getCanvasHeight(self):
        return self.canvasHeight

    def read(self, br):
        self.paramDefSet = br.readObject()
        self.partsDataList = br.readObject()
        self.canvasWidth = br.readInt32()
        self.canvasHeight = br.readInt32()

    def getPartsDataList(self):
        return self.partsDataList

    def getParamDefSet(self):
        return self.paramDefSet



from ..motion import MotionQueueEntry


class MotionQueueManager:

    def __init__(self):
        self.motions = []

    def startMotion(self, aJ, aI):
        count = len(self.motions)
        for i in range(0, count, 1):
            ent = self.motions[i]
            if ent is None:
                continue

            ent.startFadeOut(ent.motion.getFadeOut())

        if aJ is None:
            return -1

        ent = MotionQueueEntry()
        ent.motion = aJ
        self.motions.append(ent)
        nr = ent.mqNo

        return nr

    def updateParam(self, aJ):
        updated = False
        i = 0
        size = len(self.motions)
        while i < size:
            ent = self.motions[i]
            if ent is None:
                self.motions.pop(i)
                size -= 1
                continue

            mtn = ent.motion
            if mtn is None:
                self.motions.pop(i)
                size -= 1
                continue

            mtn.updateParam(aJ, ent)
            updated = True
            if ent.isFinished():
                ent = self.motions.pop(i)
                size -= 1
                i -= 1
            else:
                pass
            i += 1

        return updated

    def isFinished(self, nr=None):
        if nr is not None:  # is the motion finished?
            for i in range(0, len(self.motions), 1):
                ent = self.motions[i]
                if ent is None:
                    continue

                if ent.mqNo == nr and not ent.isFinished():
                    return False

            return True
        else:  # are all the motions finished?
            i = 0
            size = len(self.motions)
            while i < size:
                ent = self.motions[i]
                if ent is None:
                    self.motions.pop(i)
                    size -= 1
                    continue

                aH = ent.motion
                if aH is None:
                    self.motions.pop(i)
                    size -= 1
                    continue

                if not ent.isFinished():
                    return False
                i += 1

            return True

    def stopAllMotions(self):
        size = len(self.motions)
        i = 0
        while i < size:
            ent = self.motions[i]
            if ent is None:
                self.motions.pop(i)
                size -= 1
                continue

            mtn = ent.motion
            if mtn is None:
                self.motions.pop(i)
                size -= 1
                continue

            self.motions.pop(i)
            size -= 1



import os
import platform
import re
import subprocess
import sys

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
from setuptools.command.bdist_wheel import bdist_wheel

NAME = "live2d-py"
VERSION = "0.4.5.1"  # TODO: edit before push
DESCRIPTION = "Live2D Python SDK"
LONG_DESCRIPTION = open("README.md", "r", encoding="utf-8").read()
AUTHOR = "Arkueid"
AUTHOR_EMAIL = "thetardis@qq.com"
URL = "https://github.com/Arkueid/live2d-py"
REQUIRES_PYTHON = ">=3.2"
INSTALL_REQUIRES = ["numpy", "pyopengl", "pillow"]


def is_virtualenv():
    return 'VIRTUAL_ENV' in os.environ


def get_base_python_path(venv_path):
    return re.search("home = (.*)\n", open(os.path.join(venv_path, "pyvenv.cfg"), 'r').read()).group(1)


cmake_built = False
def run_cmake():
    global cmake_built
    if cmake_built: return

    cmake_args = []
    build_args = ["--config", "Release"]

    if platform.system() == "Windows":
        if platform.python_compiler().find("64 bit") > 0:
            print("Building for 64 bit")
            cmake_args += ["-A", "x64"]
        else:
            print("Building for 32 bit")
            cmake_args += ["-A", "Win32"]
        # native options
        build_args += ["--", "/m:2"]
    else:
        cmake_args += ["-DCMAKE_BUILD_TYPE=" + "Release"]
        build_args += ["--", "-j2"]
    build_folder = os.path.join(os.getcwd(), "build")

    if not os.path.exists(build_folder):
        os.makedirs(build_folder)

    if is_virtualenv():
        python_installation_path = get_base_python_path(os.environ["VIRTUAL_ENV"])
    else:
        python_installation_path = os.path.split(sys.executable)[0]
    print("Python installation path: " + python_installation_path)
    sys.stdout.flush()

    cmake_args += ["-DPYTHON_INSTALLATION_PATH=" + python_installation_path]

    cmake_setup = ["cmake", ".."] + cmake_args
    cmake_build = ["cmake", "--build", "."] + build_args

    print("Building extension for Python {}".format(sys.version.split('\n', 1)[0]))
    print("Invoking CMake setup: '{}'".format(' '.join(cmake_setup)))
    sys.stdout.flush()
    subprocess.check_call(cmake_setup, cwd=build_folder)
    print("Invoking CMake build: '{}'".format(' '.join(cmake_build)))
    sys.stdout.flush()
    subprocess.check_call(cmake_build, cwd=build_folder)

    cmake_built = True


class FakeExtension(Extension):

    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[], py_limited_api=True)
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):

    def run(self):
        run_cmake()


class BuildWheel(bdist_wheel):
    def run(self):
        run_cmake()
        bdist_wheel.run(self)


class Install(install):
    def run(self):
        run_cmake()
        install.run(self)


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license="LICENSE",
    url=URL,
    install_requires=INSTALL_REQUIRES,
    ext_modules=[FakeExtension("LAppModelWrapper", ".")],
    cmdclass={
        "build_ext": CMakeBuild,
        "bdist_wheel": BuildWheel,
        "install": Install
    },
    packages=["live2d"],
    package_data={"live2d": ["**/*.pyd", "**/*.so", "**/*.pyi", "**/*.py"]},
    package_dir={"live2d": "package/live2d"},
    keywords=["Live2D", "Cubism Live2D", "Cubism SDK", "Cubism SDK for Python"],
    python_requires=REQUIRES_PYTHON
)



from typing import TYPE_CHECKING, Optional


from .parts_context import PartsDataContext
from ..io.iserializable import ISerializable
from ..type import Array

if TYPE_CHECKING:
    from ..draw import Mesh
    from ..id import Id


class PartsData(ISerializable):

    def __init__(self):
        self.visible = True
        self.locked = False
        self.id: Optional['Id'] = None
        self.deformerList = None
        self.drawDataList: list[Mesh] | None = None

    def initDirect(self):
        self.deformerList = Array()
        self.drawDataList = Array()

    def read(self, aH):
        self.locked = aH.readBit()
        self.visible = aH.readBit()
        self.id = aH.readObject()
        self.deformerList = aH.readObject()
        self.drawDataList = aH.readObject()

    def init(self):
        aH = PartsDataContext(self)
        aH.setPartsOpacity(1 if self.isVisible() else 0)
        return aH

    def setDeformerList(self, aH):
        self.deformerList = aH

    def setDrawDataList(self, aH):
        self.drawDataList = aH

    def isVisible(self):
        return self.visible

    def isLocked(self):
        return self.locked

    def setVisible(self, aH):
        self.visible = aH

    def setLocked(self, aH):
        self.locked = aH

    def getDeformer(self):
        return self.deformerList

    def getDrawData(self):
        return self.drawDataList

    def getId(self):
        return self.id

    def setId(self, aH):
        self.id = aH



from typing import List
from typing import TYPE_CHECKING

from .ut_system import UtSystem
from ..type import Float32Array

if TYPE_CHECKING:
    from live2d.core.model_context import ModelContext
    from live2d.core.param import PivotManager


class UtInterpolate:
    @staticmethod
    def interpolateInt(mdc: 'ModelContext', pivotMgr: 'PivotManager', ret: List[bool], pivotValue: List[int]):
        a1 = pivotMgr.calcPivotValues(mdc, ret)
        a3 = mdc.getTempPivotTableIndices()
        ba = mdc.getTempT()
        pivotMgr.calcPivotIndices(a3, ba, a1)
        if a1 <= 0:
            return pivotValue[a3[0]]

        elif a1 == 1:
            bj = pivotValue[a3[0]]
            bi = pivotValue[a3[1]]
            a9 = ba[0]
            return int(bj + (bi - bj) * a9)

        elif a1 == 2:
            bj = pivotValue[a3[0]]
            bi = pivotValue[a3[1]]
            a0 = pivotValue[a3[2]]
            aZ = pivotValue[a3[3]]
            a9 = ba[0]
            a8 = ba[1]
            br = int(bj + (bi - bj) * a9)
            bq = int(a0 + (aZ - a0) * a9)
            return int(br + (bq - br) * a8)
        elif a1 == 3:
            aP = pivotValue[a3[0]]
            aO = pivotValue[a3[1]]
            bn = pivotValue[a3[2]]
            bm = pivotValue[a3[3]]
            aK = pivotValue[a3[4]]
            aJ = pivotValue[a3[5]]
            bg = pivotValue[a3[6]]
            bf = pivotValue[a3[7]]
            a9 = ba[0]
            a8 = ba[1]
            a6 = ba[2]
            bj = int(aP + (aO - aP) * a9)
            bi = int(bn + (bm - bn) * a9)
            a0 = int(aK + (aJ - aK) * a9)
            aZ = int(bg + (bf - bg) * a9)
            br = int(bj + (bi - bj) * a8)
            bq = int(a0 + (aZ - a0) * a8)
            return int(br + (bq - br) * a6)
        elif a1 == 4:
            aT = pivotValue[a3[0]]
            aS = pivotValue[a3[1]]
            bu = pivotValue[a3[2]]
            bt = pivotValue[a3[3]]
            aN = pivotValue[a3[4]]
            aM = pivotValue[a3[5]]
            bl = pivotValue[a3[6]]
            bk = pivotValue[a3[7]]
            be = pivotValue[a3[8]]
            bc = pivotValue[a3[9]]
            aX = pivotValue[a3[10]]
            aW = pivotValue[a3[11]]
            a7 = pivotValue[a3[12]]
            a5 = pivotValue[a3[13]]
            aR = pivotValue[a3[14]]
            aQ = pivotValue[a3[15]]
            a9 = ba[0]
            a8 = ba[1]
            a6 = ba[2]
            a4 = ba[3]
            aP = int(aT + (aS - aT) * a9)
            aO = int(bu + (bt - bu) * a9)
            bn = int(aN + (aM - aN) * a9)
            bm = int(bl + (bk - bl) * a9)
            aK = int(be + (bc - be) * a9)
            aJ = int(aX + (aW - aX) * a9)
            bg = int(a7 + (a5 - a7) * a9)
            bf = int(aR + (aQ - aR) * a9)
            bj = int(aP + (aO - aP) * a8)
            bi = int(bn + (bm - bn) * a8)
            a0 = int(aK + (aJ - aK) * a8)
            aZ = int(bg + (bf - bg) * a8)
            br = int(bj + (bi - bj) * a6)
            bq = int(a0 + (aZ - a0) * a6)
            return int(br + (bq - br) * a4)
        else:
            aV = 1 << a1
            aY = Float32Array(aV)
            for bh in range(0, aV, 1):
                aI: float = bh
                aH = 1
                for aL in range(0, a1, 1):
                    aH *= (1 - ba[aL]) if (aI % 2 == 0) else ba[aL]
                    aI /= 2

                aY[bh] = aH

            bs = Float32Array(aV)
            for aU in range(0, aV, 1):
                bs[aU] = pivotValue[a3[aU]]

            bd = 0
            for aU in range(0, aV, 1):
                bd += aY[aU] * bs[aU]

            return int(bd + 0.5)

    @staticmethod
    def interpolateFloat(mdc: 'ModelContext', pivotMgr: 'PivotManager', ret: List[bool], pivotValue: List[float]):
        a1 = pivotMgr.calcPivotValues(mdc, ret)
        a2 = mdc.getTempPivotTableIndices()
        a9 = mdc.getTempT()
        pivotMgr.calcPivotIndices(a2, a9, a1)
        if a1 <= 0:
            return pivotValue[a2[0]]
        if a1 == 1:
            bj = pivotValue[a2[0]]
            bi = pivotValue[a2[1]]
            a8 = a9[0]
            return bj + (bi - bj) * a8
        elif a1 == 2:
            bj = pivotValue[a2[0]]
            bi = pivotValue[a2[1]]
            a0 = pivotValue[a2[2]]
            aZ = pivotValue[a2[3]]
            a8 = a9[0]
            a7 = a9[1]
            return (1 - a7) * (bj + (bi - bj) * a8) + a7 * (a0 + (aZ - a0) * a8)
        elif a1 == 3:
            aP = pivotValue[a2[0]]
            aO = pivotValue[a2[1]]
            bn = pivotValue[a2[2]]
            bm = pivotValue[a2[3]]
            aK = pivotValue[a2[4]]
            aJ = pivotValue[a2[5]]
            bf = pivotValue[a2[6]]
            be = pivotValue[a2[7]]
            a8 = a9[0]
            a7 = a9[1]
            a5 = a9[2]
            return (1 - a5) * ((1 - a7) * (aP + (aO - aP) * a8) + a7 * (bn + (bm - bn) * a8)) + a5 * (
                    (1 - a7) * (aK + (aJ - aK) * a8) + a7 * (bf + (be - bf) * a8))
        elif a1 == 4:
            aT = pivotValue[a2[0]]
            aS = pivotValue[a2[1]]
            bs = pivotValue[a2[2]]
            br = pivotValue[a2[3]]
            aN = pivotValue[a2[4]]
            aM = pivotValue[a2[5]]
            bl = pivotValue[a2[6]]
            bk = pivotValue[a2[7]]
            bd = pivotValue[a2[8]]
            bb = pivotValue[a2[9]]
            aX = pivotValue[a2[10]]
            aW = pivotValue[a2[11]]
            a6 = pivotValue[a2[12]]
            a4 = pivotValue[a2[13]]
            aR = pivotValue[a2[14]]
            aQ = pivotValue[a2[15]]
            a8 = a9[0]
            a7 = a9[1]
            a5 = a9[2]
            a3 = a9[3]
            return (1 - a3) * ((1 - a5) * (
                    (1 - a7) * (aT + (aS - aT) * a8) + a7 * (bs + (br - bs) * a8)) + a5 * (
                                       (1 - a7) * (aN + (aM - aN) * a8) + a7 * (
                                       bl + (bk - bl) * a8))) + a3 * ((1 - a5) * (
                    (1 - a7) * (bd + (bb - bd) * a8) + a7 * (aX + (aW - aX) * a8)) + a5 * (
                                                                              (1 - a7) * (
                                                                              a6 + (
                                                                              a4 - a6) * a8) + a7 * (
                                                                                      aR + (
                                                                                      aQ - aR) * a8)))
        else:
            aV = 1 << a1
            aY = Float32Array(aV)
            for bh in range(0, aV, 1):
                aI: float = bh
                aH = 1
                for aL in range(0, a1, 1):
                    aH *= (1 - a9[aL]) if (aI % 2 == 0) else a9[aL]
                    aI /= 2

                aY[bh] = aH

            bq = Float32Array(aV)
            for aU in range(0, aV, 1):
                bq[aU] = pivotValue[a2[aU]]

            bc = 0
            for aU in range(0, aV, 1):
                bc += aY[aU] * bq[aU]

            return bc

    @staticmethod
    def interpolatePoints(mdc: 'ModelContext', pivotMgr: 'PivotManager', retParamOut: List[bool], numPts: int, pivotPoints: List[float], dstPoints, ptOffset, ptStep):
        aN = pivotMgr.calcPivotValues(mdc, retParamOut)
        bw = mdc.getTempPivotTableIndices()
        a2 = mdc.getTempT()
        pivotMgr.calcPivotIndices(bw, a2, aN)
        aJ = numPts * 2
        aQ = ptOffset
        if aN <= 0:
            bI = bw[0]
            bq = pivotPoints[bI]
            if ptStep == 2 and ptOffset == 0:
                UtSystem.arraycopy(bq, 0, dstPoints, 0, aJ)
            else:
                bt = 0
                while bt < aJ:
                    dstPoints[aQ] = bq[bt]
                    bt += 1
                    dstPoints[aQ + 1] = bq[bt]
                    bt += 1
                    aQ += ptStep
        elif aN == 1:
            bq = pivotPoints[bw[0]]
            bp = pivotPoints[bw[1]]
            b3 = a2[0]
            bT = 1 - b3
            bt = 0
            while bt < aJ:
                dstPoints[aQ] = bq[bt] * bT + bp[bt] * b3
                bt += 1
                dstPoints[aQ + 1] = bq[bt] * bT + bp[bt] * b3
                bt += 1
                aQ += ptStep
        elif aN == 2:
            bq = pivotPoints[bw[0]]
            bp = pivotPoints[bw[1]]
            aZ = pivotPoints[bw[2]]
            aY = pivotPoints[bw[3]]
            b3 = a2[0]
            b1 = a2[1]
            bT = 1 - b3
            bP = 1 - b1
            b2 = bP * bT
            b0 = bP * b3
            bM = b1 * bT
            bL = b1 * b3
            bt = 0
            while bt < aJ:
                dstPoints[aQ] = b2 * bq[bt] + b0 * bp[bt] + bM * aZ[bt] + bL * aY[bt]
                bt += 1
                dstPoints[aQ + 1] = b2 * bq[bt] + b0 * bp[bt] + bM * aZ[bt] + bL * aY[bt]
                bt += 1
                aQ += ptStep
        elif aN == 3:
            ba = pivotPoints[bw[0]]
            a9 = pivotPoints[bw[1]]
            aP = pivotPoints[bw[2]]
            aO = pivotPoints[bw[3]]
            a6 = pivotPoints[bw[4]]
            a4 = pivotPoints[bw[5]]
            aL = pivotPoints[bw[6]]
            aK = pivotPoints[bw[7]]
            b3 = a2[0]
            b1 = a2[1]
            bZ = a2[2]
            bT = 1 - b3
            bP = 1 - b1
            bN = 1 - bZ
            b8 = bN * bP * bT
            b7 = bN * bP * b3
            bU = bN * b1 * bT
            bS = bN * b1 * b3
            b6 = bZ * bP * bT
            b5 = bZ * bP * b3
            bQ = bZ * b1 * bT
            bO = bZ * b1 * b3
            bt = 0
            while bt < aJ:
                dstPoints[aQ] = b8 * ba[bt] + b7 * a9[bt] + bU * aP[bt] + bS * aO[bt] + b6 * a6[bt] + b5 * a4[
                    bt] + bQ * aL[bt] + bO * aK[bt]
                bt += 1
                dstPoints[aQ + 1] = b8 * ba[bt] + b7 * a9[bt] + bU * aP[bt] + bS * aO[bt] + b6 * a6[bt] + b5 * a4[
                    bt] + bQ * aL[bt] + bO * aK[bt]
                bt += 1
                aQ += ptStep
        elif aN == 4:
            bD = pivotPoints[bw[0]]
            bB = pivotPoints[bw[1]]
            bo = pivotPoints[bw[2]]
            bm = pivotPoints[bw[3]]
            by = pivotPoints[bw[4]]
            bx = pivotPoints[bw[5]]
            be = pivotPoints[bw[6]]
            bd = pivotPoints[bw[7]]
            bG = pivotPoints[bw[8]]
            bE = pivotPoints[bw[9]]
            bv = pivotPoints[bw[10]]
            bu = pivotPoints[bw[11]]
            bA = pivotPoints[bw[12]]
            bz = pivotPoints[bw[13]]
            bn = pivotPoints[bw[14]]
            bl = pivotPoints[bw[15]]
            b3 = a2[0]
            b1 = a2[1]
            bZ = a2[2]
            bY = a2[3]
            bT = 1 - b3
            bP = 1 - b1
            bN = 1 - bZ
            bK = 1 - bY
            bk = bK * bN * bP * bT
            bi = bK * bN * bP * b3
            aW = bK * bN * b1 * bT
            aV = bK * bN * b1 * b3
            bc = bK * bZ * bP * bT
            bb = bK * bZ * bP * b3
            aS = bK * bZ * b1 * bT
            aR = bK * bZ * b1 * b3
            bs = bY * bN * bP * bT
            br = bY * bN * bP * b3
            a1 = bY * bN * b1 * bT
            a0 = bY * bN * b1 * b3
            bh = bY * bZ * bP * bT
            bf = bY * bZ * bP * b3
            aU = bY * bZ * b1 * bT
            aT = bY * bZ * b1 * b3
            bt = 0
            while bt < aJ:
                dstPoints[aQ] = bk * bD[bt] + bi * bB[bt] + aW * bo[bt] + aV * bm[bt] + bc * by[bt] + bb * bx[
                    bt] + aS * be[bt] + aR * bd[bt] + bs * bG[bt] + br * bE[bt] + a1 * bv[bt] + a0 * bu[
                                    bt] + bh * bA[bt] + bf * bz[bt] + aU * bn[bt] + aT * bl[bt]
                bt += 1
                dstPoints[aQ + 1] = bk * bD[bt] + bi * bB[bt] + aW * bo[bt] + aV * bm[bt] + bc * by[bt] + bb * \
                                    bx[bt] + aS * be[bt] + aR * bd[bt] + bs * bG[bt] + br * bE[bt] + a1 * bv[
                                        bt] + a0 * bu[bt] + bh * bA[bt] + bf * bz[bt] + aU * bn[bt] + aT * bl[
                                        bt]
                bt += 1
                aQ += ptStep
        else:
            b4 = 1 << aN
            bJ = Float32Array(b4)
            for bj in range(0, b4, 1):
                aH: float = bj
                aM = 1
                for bF in range(0, aN, 1):
                    aM *= (1 - a2[bF]) if (aH % 2 == 0) else a2[bF]
                    aH /= 2

                bJ[bj] = aM

            bg = Float32Array(b4)
            for aX in range(0, b4, 1):
                bg[aX] = pivotPoints[bw[aX]]

            bt = 0
            while bt < aJ:
                a8 = 0
                a7 = 0
                bR = bt + 1
                for aX in range(0, b4, 1):
                    a8 += bJ[aX] * bg[aX][bt]
                    a7 += bJ[aX] * bg[aX][bR]

                bt += 2
                dstPoints[aQ] = a8
                dstPoints[aQ + 1] = a7
                aQ += ptStep



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from mesh import Mesh

class IDrawContext:

    def __init__(self, dd: 'Mesh'):
        self.interpolatedDrawOrder: Optional[int] = None
        self.paramOutside: bool = False
        self.partsOpacity: float = 0
        self.available: bool = True
        self.baseOpacity: float = 1.0
        self.clipBufPre_clipContext = None
        self.drawData: 'Mesh' = dd
        self.partsIndex = -1

    def isParamOutside(self):
        return self.paramOutside

    def isAvailable(self):
        return self.available and not self.paramOutside

    def getDrawData(self) -> 'Mesh':
        return self.drawData



class TextureInfo:

    def __init__(self):
        self.a: float = 1.
        self.r: float = 1.
        self.g: float = 1.
        self.b: float = 1.
        # self.scale: float = 1.
        # self.interpolate = 1
        # self.blendMode: int = 0



from typing import TYPE_CHECKING, Union

from ..io.iserializable import ISerializable

if TYPE_CHECKING:
    from ..id import Id


class ParamPivots(ISerializable):
    PARAM_INDEX_NOT_INIT = -2

    def __init__(self):
        self.pivotCount: int = 0
        self.paramId: 'Id' | None = None
        self.pivotValues: list[float] | None = None
        self.paramIndex: int = ParamPivots.PARAM_INDEX_NOT_INIT
        self.initVersion: int = -1
        self.tmpPivotIndex: int = 0
        self.tmpT: float = 0

    def read(self, br):
        self.paramId = br.readObject()
        self.pivotCount = br.readInt32()
        self.pivotValues = br.readObject()

    def getParamIndex(self, initVersion):
        if self.initVersion != initVersion:
            self.paramIndex = ParamPivots.PARAM_INDEX_NOT_INIT

        return self.paramIndex

    def setParamIndex(self, index: int, initVersion: int):
        self.paramIndex = index
        self.initVersion = initVersion

    def getParamID(self) -> 'Id':
        return self.paramId

    def getPivotCount(self):
        return self.pivotCount

    def getPivotValues(self):
        return self.pivotValues

    def getTmpPivotIndex(self):
        return self.tmpPivotIndex

    def setTmpPivotIndex(self, index):
        self.tmpPivotIndex = index

    def getTmpT(self) -> float:
        return self.tmpT

    def setTmpT(self, value: float):
        self.tmpT = value



class L2DMatrix44:

    def __init__(self):
        self.tr = [0] * 16
        self.identity()

    def identity(self):
        for i in range(16):
            self.tr[i] = 1 if ((i % 5) == 0) else 0

    def getArray(self):
        return self.tr

    def getCopyMatrix(self):
        return self.tr.copy()

    def setMatrix(self, tr):
        if self.tr is None or self.tr.length != self.tr.length:
            return
        for i in range(16):
            self.tr[i] = tr[i]

    def getScaleX(self):
        return self.tr[0]

    def getScaleY(self):
        return self.tr[5]

    def transformX(self, src):
        return self.tr[0] * src + self.tr[12]

    def transformY(self, src):
        return self.tr[5] * src + self.tr[13]

    def invertTransformX(self, src):
        return (src - self.tr[12]) / self.tr[0]

    def invertTransformY(self, src):
        return (src - self.tr[13]) / self.tr[5]

    def multTranslate(self, shiftX, shiftY):
        tr1 = [1, 0, 0, 0,
               0, 1, 0, 0,
               0, 0, 1, 0,
               shiftX, shiftY, 0, 1]
        L2DMatrix44.mul(tr1, self.tr, self.tr)

    def translate(self, x, y):
        self.tr[12] = x
        self.tr[13] = y

    def translateX(self, x):
        self.tr[12] = x

    def translateY(self, y):
        self.tr[13] = y

    def multScale(self, scaleX, scaleY):
        tr1 = [scaleX, 0, 0, 0, 0, scaleY, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]
        L2DMatrix44.mul(tr1, self.tr, self.tr)

    def scale(self, scaleX, scaleY):
        self.tr[0] = scaleX
        self.tr[5] = scaleY

    @staticmethod
    def mul(a, b, dst):
        c = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        n = 4
        for i in range(n):
            for j in range(n):
                for k in range(n):
                    c[i + j * 4] += a[i + k * 4] * b[k + j * 4]

        for i in range(16):
            dst[i] = c[i]



from ...core import MotionQueueManager


class L2DMotionManager(MotionQueueManager):

    def __init__(self):
        super().__init__()
        self.currentPriority = 0
        self.reservePriority = 0

    def getCurrentPriority(self):
        return self.currentPriority

    def getReservePriority(self):
        return self.reservePriority

    def reserveMotion(self, priority):
        if self.reservePriority >= priority:
            return False

        if self.currentPriority >= priority:
            return False

        self.reservePriority = priority
        return True

    def setReservePriority(self, val):
        self.reservePriority = val

    def updateParam(self, model):
        updated = super().updateParam(model)
        if self.isFinished():
            self.currentPriority = 0

        return updated

    def startMotionPrio(self, motion, priority):
        if priority == self.reservePriority:
            self.reservePriority = 0

        self.currentPriority = priority
        return self.startMotion(motion, False)



from ...core import Id


class L2DPartsParam:

    def __init__(self, ppid):
        self.paramIndex = -1
        self.partsIndex = -1
        self.link = None
        self.id = ppid

    def initIndex(self, model):
        self.paramIndex = model.getParamIndex("VISIBLE:" + self.id)
        self.partsIndex = model.getPartsDataIndex(Id.getID(self.id))
        model.setParamFloat(self.paramIndex, 1)



from ..util import UtSystem


class MotionQueueEntry:
    MQ_NO = 0

    def __init__(self):
        self.motion = None
        self.available = True
        self.finished = False
        self.startTimeMSec = -1
        self.fadeInStartTimeMSec = -1
        self.endTimeMSec = -1
        self.mqNo = MotionQueueEntry.MQ_NO
        MotionQueueEntry.MQ_NO += 1

    def isFinished(self):
        return self.finished

    def startFadeOut(self, fadeOutMSec):
        ct = UtSystem.getUserTimeMSec()
        new_end_time_m_sec = ct + fadeOutMSec
        if self.endTimeMSec < 0 or new_end_time_m_sec < self.endTimeMSec:
            self.endTimeMSec = new_end_time_m_sec
